# **************************************************************************
# * Authors:    Mohamad Harastani            (mohamad.harastani@igbmc.fr)
# *             Slavica Jonic                (slavica.jonic@upmc.fr)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os
from os.path import basename, join, exists, isfile
import numpy as np
import pwem.emlib.metadata as md
from pyworkflow.utils.path import cleanPath, makePath
from pyworkflow.viewer import (ProtocolViewer, DESKTOP_TKINTER, WEB_DJANGO)
from pyworkflow.protocol.params import StringParam, LabelParam
from pwem.objects import SetOfParticles
from pyworkflow.gui.browser import FileBrowserWindow
from continuousflex.protocols.protocol_tomoflow_dimred import FlexProtDimredHeteroFlow
from continuousflex.protocols.data import Point, Data
from .plotter_vol import FlexNmaVolPlotter
from continuousflex.viewers.nma_vol_gui import ClusteringWindowVolHeteroFlow
from continuousflex.viewers.nma_vol_gui import TrajectoriesWindowVolHeteroFlow
from pwem.viewers.viewer_chimera import Chimera
from joblib import load, dump
from continuousflex.protocols.utilities.spider_files3 import open_volume, save_volume
import continuousflex
from continuousflex import Plugin
from subprocess import check_call
import sys
import matplotlib.pyplot as plt
from pwem.emlib.image import ImageHandler
from pyworkflow.protocol import params

FIGURE_LIMIT_NONE = 0
FIGURE_LIMITS = 1

X_LIMITS_NONE = 0
X_LIMITS = 1
Y_LIMITS_NONE = 0
Y_LIMITS = 1
Z_LIMITS_NONE = 0
Z_LIMITS = 1
POINT_LIMITS_NONE = 0
POINT_LIMITS = 1


class FlexDimredHeteroFlowViewer(ProtocolViewer):
    """ Visualization of results from TomoFlow Dimred protocol
    """
    _label = 'viewer heteroflow dimred'
    _targets = [FlexProtDimredHeteroFlow]
    _environments = [DESKTOP_TKINTER, WEB_DJANGO]

    def __init__(self, **kwargs):
        ProtocolViewer.__init__(self, **kwargs)
        self._data = None

    def getData(self):
        if self._data is None:
            self._data = self.loadData()
        return self._data

    def _defineParams(self, form):
        form.addSection(label='Visualization')
        form.addParam('displayRawDeformation', StringParam, default='1 2',
                      label='Display raw deformation',
                      help='Type 1 to see the histogram of reduced dimensions, '
                           'using axis 1; \n '
                           'Type 2 to see the histogram of reduced dimensions, '
                           'using axis 2; etc. \n '
                           'Type 1 2 to see the histogram of reduced dimensions, using axes 1 and 2; \n'
                           'Type 1 2 3 to see the histogram of reduced dimensions, using axes 1, 2, '
                           'and 3; etc. '
                      )
        form.addParam('displayPcaSingularValues', LabelParam,
                      label="Display PCA singular values",
                      help="The values should help you see how many dimensions are in the data ")
        form.addParam('displayClustering', LabelParam,
                      label='Open clustering tool?',
                      help='Open a GUI to visualize the volumes as points '
                           'and select some of them to create new clusters, and compute 3D averages of the clusters')
        form.addParam('displayTrajectories', LabelParam,
                      label='Open trajectories tool?',
                      help='Open a GUI to visualize the volumes as points'
                           ' to draw and adjust trajectories.')
        form.addParam('graylevel',params.FloatParam, label='Gray-level threshold level for animations',
                      default=0.1, expertLevel=params.LEVEL_ADVANCED)
        form.addHidden('limits_modes', params.EnumParam,
                      choices=['Automatic (Recommended)', 'Set manually Use upper and lower values'],
                      default=FIGURE_LIMIT_NONE,
                      label='(1 - CC) limits', display=params.EnumParam.DISPLAY_COMBO,
                      help='If you want to use a range of (1-CC) choose to set it manually.')
        form.addHidden('LimitLow', params.FloatParam, default=None,
                      condition='limits_modes==%d' % FIGURE_LIMITS,
                      label='Lower (1-CC) value',
                      help='The lower (1-CC) used in the graph')
        form.addHidden('LimitHigh', params.FloatParam, default=None,
                      condition='limits_modes==%d' % FIGURE_LIMITS,
                      label='Upper (1-CC) value',
                      help='The upper (1-CC) used in the graph')
        form.addParam('xlimits_mode', params.EnumParam,
                      choices=['Automatic (Recommended)', 'Set manually x-axis limits'],
                      default=X_LIMITS_NONE,
                      label='x-axis limits', display=params.EnumParam.DISPLAY_COMBO,
                      help='This allows you to use a specific range of x-axis limits')
        form.addParam('xlim_low', params.FloatParam, default=None,
                      condition='xlimits_mode==%d' % X_LIMITS,
                      label='Lower x-axis limit')
        form.addParam('xlim_high', params.FloatParam, default=None,
                      condition='xlimits_mode==%d' % X_LIMITS,
                      label='Upper x-axis limit')
        form.addParam('ylimits_mode', params.EnumParam,
                      choices=['Automatic (Recommended)', 'Set manually y-axis limits'],
                      default=Y_LIMITS_NONE,
                      label='y-axis limits', display=params.EnumParam.DISPLAY_COMBO,
                      help='This allows you to use a specific range of y-axis limits')
        form.addParam('ylim_low', params.FloatParam, default=None,
                      condition='ylimits_mode==%d' % Y_LIMITS,
                      label='Lower y-axis limit')
        form.addParam('ylim_high', params.FloatParam, default=None,
                      condition='ylimits_mode==%d' % Y_LIMITS,
                      label='Upper y-axis limit')
        form.addParam('zlimits_mode', params.EnumParam,
                      choices=['Automatic (Recommended)', 'Set manually z-axis limits'],
                      default=Z_LIMITS_NONE,
                      label='z-axis limits', display=params.EnumParam.DISPLAY_COMBO,
                      help='This allows you to use a specific range of z-axis limits')
        form.addParam('zlim_low', params.FloatParam, default=None,
                      condition='zlimits_mode==%d' % Z_LIMITS,
                      label='Lower z-axis limit')
        form.addParam('zlim_high', params.FloatParam, default=None,
                      condition='zlimits_mode==%d' % Z_LIMITS,
                      label='Upper z-axis limit')
        # Scatter points size and transparancy
        form.addParam('points_shades', params.EnumParam,
                      choices=['Automatic (Recommended)', 'Set manually point radius and transparancy'],
                      default=POINT_LIMITS_NONE,
                      label='Scatter points radius and transparancy', display=params.EnumParam.DISPLAY_COMBO,
                      help='This allows you to use change the points radius and transparancy in the scatter plot'
                           '. By trying different values, it may help you discover the densest regions in the space.')
        line = form.addLine('Radius and transparancy',
                            condition='points_shades==%d' % POINT_LIMITS,
                            help='Values for points rarius have can be any positive real number.'
                                 ' Values for transparancy are between 0 and 1.')
        line.addParam('s', params.FloatParam, default=None, allowsNull=True,
                      label='Radius')
        line.addParam('alpha', params.FloatParam, default=None, allowsNull=True,
                        label='Transparancy')


    def _getVisualizeDict(self):
        return {'displayRawDeformation': self._viewRawDeformation,
                'displayPcaSingularValues': self.viewPcaSinglularValues,
                'displayClustering': self._displayClustering,
                'displayTrajectories': self._displayTrajectories,
                }

    def _viewRawDeformation(self, paramName):
        components = self.displayRawDeformation.get()
        return self._doViewRawDeformation(components)

    def _doViewRawDeformation(self, components):
        components = list(map(int, components.split()))
        dim = len(components)
        views = []

        if dim > 0:
            modeList = [m - 1 for m in components]
            modeNameList = ['Principle Component Axis %d' % m for m in components]
            missingList = []

            if missingList:
                return [self.errorMessage("Invalid mode(s) *%s*\n." % (', '.join(missingList)),
                                          title="Invalid input")]

            # Actually plot
            # plotter = FlexNmaVolPlotter(data=self.getData())
            if self.limits_modes == FIGURE_LIMIT_NONE:
                plotter = FlexNmaVolPlotter(data=self.getData(),
                                            xlim_low=self.xlim_low, xlim_high=self.xlim_high,
                                            ylim_low=self.ylim_low, ylim_high=self.ylim_high,
                                            zlim_low=self.zlim_low, zlim_high=self.zlim_high,
                                            s=self.s, alpha = self.alpha)
            else:
                plotter = FlexNmaVolPlotter(data=self.getData(),
                                            LimitL=self.LimitLow, LimitH=self.LimitHigh,
                                            xlim_low=self.xlim_low, xlim_high=self.xlim_high,
                                            ylim_low=self.ylim_low, ylim_high=self.ylim_high,
                                            zlim_low=self.zlim_low, zlim_high=self.zlim_high,
                                            s=self.s, alpha = self.alpha)


            baseList = [basename(n) for n in modeNameList]

            self.getData().XIND = modeList[0]
            if dim == 1:
                plotter.plotArray1D("Histogram of: %s" % baseList[0],
                                    "Amplitude", "Number of volumes")
            else:
                self.getData().YIND = modeList[1]
                if dim == 2:
                    # plotter.plotArray2D("Reduced dimensions deformation amplitudes: %s vs %s" % tuple(baseList),
                    #                     *baseList)
                    plotter.plotArray2D_xy("%s vs %s" % tuple(baseList),
                                        *baseList)
                elif dim == 3:
                    self.getData().ZIND = modeList[2]
                    # plotter.plotArray3D("%s %s %s" % tuple(baseList),
                    #                     *baseList)
                    plotter.plotArray3D_xyz("%s %s %s" % tuple(baseList),
                                        *baseList)
            views.append(plotter)

        return views

    def _displayClustering(self, paramName):
        self.clusterWindow = self.tkWindow(ClusteringWindowVolHeteroFlow,
                                           title='Volume Clustering Tool',
                                           dim=self.protocol.reducedDim.get(),
                                           data=self.getData(),
                                           callback=self._createCluster,
                                           limits_mode=self.limits_modes,
                                           LimitL=self.LimitLow,
                                           LimitH=self.LimitHigh,
                                           xlim_low=self.xlim_low,
                                           xlim_high=self.xlim_high,
                                           ylim_low=self.ylim_low,
                                           ylim_high=self.ylim_high,
                                           zlim_low=self.zlim_low,
                                           zlim_high=self.zlim_high,
                                           s=self.s,
                                           alpha=self.alpha)
        return [self.clusterWindow]

    def _displayTrajectories(self, paramName):
        self.trajectoriesWindow = self.tkWindow(TrajectoriesWindowVolHeteroFlow,
                                                title='Trajectories Tool',
                                                dim=self.protocol.reducedDim.get(),
                                                data=self.getData(),
                                                callback=self._generateAnimation,
                                                loadCallback=self._loadAnimation,
                                                numberOfPoints=10,
                                                limits_mode=self.limits_modes,
                                                LimitL=self.LimitLow,
                                                LimitH=self.LimitHigh,
                                                xlim_low=self.xlim_low,
                                                xlim_high=self.xlim_high,
                                                ylim_low=self.ylim_low,
                                                ylim_high=self.ylim_high,
                                                zlim_low=self.zlim_low,
                                                zlim_high=self.zlim_high,
                                                s=self.s,
                                                alpha=self.alpha)

        return [self.trajectoriesWindow]

    def _createCluster(self):
        """ Create the cluster with the selected particles
        from the cluster. This method will be called when
        the button 'Create Cluster' is pressed.
        """
        # Write the particles
        prot = self.protocol
        project = prot.getProject()
        # inputSet = prot.getInputParticles().get()
        inputSet = prot.getInputParticles()
        makePath(prot._getTmpPath())
        fnSqlite = prot._getTmpPath('cluster_particles.sqlite')
        cleanPath(fnSqlite)
        partSet = SetOfParticles(filename=fnSqlite)
        partSet.copyInfo(inputSet)
        first = True
        for point in self.getData():
            if point.getState() == Point.SELECTED:
                particle = inputSet[point.getId()]
                partSet.append(particle)
        partSet.write()
        partSet.close()

        from continuousflex.protocols.protocol_batch_cluster_tomoflow import FlexBatchProtTomoFlowCluster

        newProt = project.newProtocol(FlexBatchProtTomoFlowCluster)
        clusterName = self.clusterWindow.getClusterName()
        if clusterName:
            newProt.setObjLabel(clusterName)
        newProt.inputHeteroFlowDimred.set(prot)
        newProt.sqliteFile.set(fnSqlite)
        project.launchProtocol(newProt)
        project.getRunsGraph()


    #TODO
    def _loadAnimationData(self, obj):
        pass

    def _loadAnimation(self):
        prot = self.protocol
        browser = FileBrowserWindow("Select the animation folder (animation_NAME)",
                                    self.getWindow(), prot._getExtraPath(),
                                    onSelect=self._loadAnimationData)
        browser.show()

    def _generateAnimation(self):
        prot = self.protocol
        # This is not getting the file correctly, we are workingaround it:
        # projectorFile = prot.getProjectorFile()
        projectorFile = prot._getExtraPath() + '/projector.txt'
        if isfile(projectorFile):
            print('Mapping found, the animation is exact inverse of the dimensionality reduction method')
        else:
            print('Mapping not found, the animation is an estimation of reversing the dimensionality reduction method')

        animation = self.trajectoriesWindow.getAnimationName()
        animationPath = prot._getExtraPath('animation_%s' % animation)

        cleanPath(animationPath)
        makePath(animationPath)
        animationRoot = join(animationPath, 'animation_%s' % animation)
        trajectoryPoints = np.array([p.getData() for p in self.trajectoriesWindow.pathData])

        if isfile(projectorFile):
            M = np.loadtxt(projectorFile)
            if prot.getMethodName()=='sklearn_PCA':
                pca = load(prot._getExtraPath('pca_pickled.txt'))
                deformations = pca.inverse_transform(trajectoryPoints)
            else:
                # TODO: add mean
                deformations = np.dot(trajectoryPoints, np.linalg.pinv(M))
            np.savetxt(animationRoot + 'trajectory.txt', trajectoryPoints)
        else:
            Y = np.loadtxt(prot.getOutputMatrixFile())
            X = np.loadtxt(prot.getDeformationFile())
            # Find closest points in deformations
            deformations = [X[np.argmin(np.sum((Y - p) ** 2, axis=1))] for p in trajectoryPoints]

        # get the original size of the input:
        mdImgs = md.MetaData(self.protocol.inputOpFlow.get()._getExtraPath('volumes_out.xmd'))
        N = 0
        for objId in mdImgs:
            N += 1

        # reading back all optical flows
        bigmat = []
        if(isfile(self.protocol._getExtraPath('bigmat_inverse.pkl'))):
            print('bigmat_inverse.txt found')
            # bigmat_pinv = np.loadtxt(self.protocol._getExtraPath('bigmat_inverse.txt'))
            bigmat_pinv = load(self.protocol._getExtraPath('bigmat_inverse.pkl'))
        else:
            if(isfile(self.protocol._getExtraPath('bigmat.pkl'))):
                bigmat = load(self.protocol._getExtraPath('bigmat.pkl'))
            else:
                for j in range(1, N+1):
                    flowj = self.read_optical_flow_by_number(j)
                    flowj = np.reshape(flowj, [3 * np.shape(flowj)[1] * np.shape(flowj)[2] * np.shape(flowj)[3]])
                    bigmat.append(flowj)
                bigmat = np.array(bigmat)
                print('bigmat created successfully')
                # np.savetxt(self.protocol._getExtraPath('bigmat.txt'),bigmat)
                dump(bigmat,self.protocol._getExtraPath('bigmat.pkl'))
                print('bigmat.pkl saved successfully')
            bigmat_pinv = np.linalg.pinv(bigmat)
            bigmat = None  # removing it from the memory
            # np.savetxt(self.protocol._getExtraPath('bigmat_inverse.txt'),bigmat_pinv)
            dump(bigmat_pinv,self.protocol._getExtraPath('bigmat_inverse.pkl'))

        line = np.matmul(bigmat_pinv, np.transpose(deformations))
        bigmat_pinv = None # removing if from the memory
        fnref = self.protocol._getExtraPath('reference.spi')
        shape = np.shape(open_volume(fnref))
        ref = open_volume(fnref)
        # dumping the reference volume to use it in an outside script
        # creating the directory Tmp since it is usually deleted after the execution of the protocol
        if not exists(self.protocol._getTmpPath()):
            os.mkdir(self.protocol._getTmpPath())
        ref_dump = self.protocol._getTmpPath('ref_dump.pkl')
        dump(ref, ref_dump)
        for i, trash in enumerate(deformations):
            flowi = np.transpose(line[:, i])
            flowi = np.reshape(flowi, [3, shape[0], shape[1], shape[2]])
            pathi = animationRoot + str(i).zfill(3) + 'deformed_by_opflow.vol'
            flow_dump = self.protocol._getTmpPath('flow_dump.pkl')
            dump(np.float32(flowi), flow_dump)
            args = " %s %s %s" % (ref_dump, flow_dump, pathi)
            script_path = continuousflex.__path__[0] + '/protocols/utilities/optflow_warp.py'
            command = "python " + script_path + args
            command = Plugin.getContinuousFlexCmd(command)
            check_call(command, shell=True, stdout=sys.stdout, stderr=sys.stderr,
                       env=None, cwd=None)

        fn_cxc = self.protocol._getExtraPath('chimera_%s.cxc' % animation)
        # cxc_command = 'open ' + animationPath + '/*.vol vseries true\n'
        cxc_command = 'open animation_%s/*.vol vseries true\n' % animation
        # cxc_command += 'volume #1 style surface level 8.0\n'
        cxc_command += 'volume #1 style surface level %f\n' % self.graylevel
        cxc_command += 'vseries play #1 loop true maxFrameRate 5 direction oscillate'
        with open(fn_cxc, 'w') as f:
            print(cxc_command, file=f)
        command = Chimera.getHome()+'/bin/ChimeraX ' + fn_cxc
        # print(Chimera.getHome())
        os.system(command)


    def loadData(self):
        """ Iterate over the images and their deformations
        to create a Data object with theirs Points.
        """
        # particles = self.protocol.getInputParticles().get()
        particles = self.protocol.getInputParticles()
        mat = np.loadtxt(self.protocol._getExtraPath('output_matrix.txt'))
        data = Data()
        for i, particle in enumerate(particles):
            pointData = mat[i,:]
            data.addPoint(Point(pointId=particle.getObjId(),
                                data=pointData,
                                weight=0))
        return data

    def _validate(self):
        errors = []
        return errors

    def read_optical_flow(self, path_flowx, path_flowy, path_flowz):
        # x = open_volume(path_flowx)
        x = ImageHandler().read(path_flowx).getData()
        # y = open_volume(path_flowy)
        y = ImageHandler().read(path_flowy).getData()
        # z = open_volume(path_flowz)
        z = ImageHandler().read(path_flowz).getData()

        l = np.shape(x)
        flow = np.zeros([3, l[0], l[1], l[2]])
        flow[0, :, :, :] = x
        flow[1, :, :, :] = y
        flow[2, :, :, :] = z
        return flow

    def read_optical_flow_by_number(self, num):
        op_path = self.protocol.inputOpFlow.get()._getExtraPath()+'/optical_flows/'
        path_flowx = op_path + str(num).zfill(6) + '_opflowx.spi'
        path_flowy = op_path + str(num).zfill(6) + '_opflowy.spi'
        path_flowz = op_path + str(num).zfill(6) + '_opflowz.spi'
        flow = self.read_optical_flow(path_flowx, path_flowy, path_flowz)
        return flow

    def viewPcaSinglularValues(self, paramName):
        pca = load(self.protocol._getExtraPath('pca_pickled.txt'))
        fig = plt.figure('PCA singlular values')
        plt.stem(pca.singular_values_)
        plt.xticks(np.arange(0, len(pca.singular_values_), 1))
        plt.show()
        pass



import os
import glob
import numpy as np

def edit(path):
    with open(path, 'r') as f:
        lines = f.readlines()
        for i in range(len(lines)):
            if i < 13:
                pass
            else:
                lines[i] = lines[i].replace(lines[i][0:72], 'synthimages_set/img'+lines[i][0:6]+'.spi')
        with open('../images.xmd', 'w') as output:
            for i in range(len(lines)):
                output.write(lines[i])
            output.close()



edit('../images.xmd')



# **************************************************************************
# *
# * Authors: Rémi Vuillemot             (remi.vuillemot@upmc.fr)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from continuousflex.viewers.nma_gui import TrajectoriesWindow, ClusteringWindow
import tkinter as tk
from pyworkflow.gui.widgets import Button, ComboBox
from tkinter import Radiobutton
import numpy as np
import scipy as sp
from continuousflex.protocols.data import Point
from sklearn.cluster import KMeans

TOOL_TRAJECTORY = 1
TOOL_CLUSTERING = 2


class PCAWindowDimred(TrajectoriesWindow, ClusteringWindow):

    def __init__(self, **kwargs):
        TrajectoriesWindow.__init__(self, **kwargs)
        self.saveClusterCallback = kwargs.get('saveClusterCallback', None)
        self.saveCallback = kwargs.get('saveCallback', None)
        self.numberOfPoints = kwargs.get('numberOfPoints', 10)

        self._alpha=self.alpha
        self._s=self.s
        self._clusterNumber = 0

    def _createContent(self, content):
        self._createModeBox(content)
        self._createFigureBox(content)
        self._createTrajectoriesBox(content)
        self._createClusteringBox(content)
        self._exportBox(content)

    def _createModeBox(self, content):
        frame = tk.LabelFrame(content, text='Interactive mode', font=self.fontBold)

        selFrame = tk.Frame(frame)
        self.selectTool = tk.IntVar()
        r1 = Radiobutton(selFrame, text="Trajectory Mode", variable=self.selectTool, value=TOOL_TRAJECTORY, command=self._onUpdateClick)
        r1.grid(row=0, column=0, padx=5)
        r2 = Radiobutton(selFrame, text="Selection Mode", variable=self.selectTool, value=TOOL_CLUSTERING, command=self._onUpdateClick)
        r2.grid(row=0, column=1, padx=5)
        self.selectTool.set(TOOL_TRAJECTORY)
        selFrame.grid(row=0, column=0)
        frame.grid(row=0, column=0, sticky='new', padx=5, pady=(10, 5))


    def _createFigureBox(self, content):
        frame = tk.LabelFrame(content, text='Figure', font=self.fontBold)
        frame.columnconfigure(0, minsize=50)
        frame.columnconfigure(1, weight=1)  # , minsize=30)
        # Create the 'Axes' label
        self._addLabel(frame, 'Axes', 0, 0)

        # Create a listbox with x1, x2 ...
        listbox = tk.Listbox(frame, height=5,
                             selectmode=tk.MULTIPLE, bg='white')
        for x in range(1, self.dim + 1):
            listbox.insert(tk.END, 'x%d' % x)
        listbox.grid(row=0, column=1, padx=5, pady=5, sticky='w')
        self.listbox = listbox

        # Selection controls
        self._addLabel(frame, 'Rejection', 1, 0)
        # Selection label
        self.selectionVar = tk.StringVar()
        self.clusterLabel = tk.Label(frame, textvariable=self.selectionVar)
        self.clusterLabel.grid(row=1, column=1, sticky='w', padx=5, pady=(10, 5))
        self._updateSelectionLabel()
        # --- Expression
        expressionFrame = tk.Frame(frame)
        expressionFrame.grid(row=2, column=1, sticky='w')
        tk.Label(expressionFrame, text='Expression').grid(row=0, column=0, sticky='ne')
        self.expressionVar = tk.StringVar()
        expressionEntry = tk.Entry(expressionFrame, textvariable=self.expressionVar,
                                   width=30, bg='white')
        expressionEntry.grid(row=0, column=1, sticky='nw')
        helpText = 'e.g. x1>0 and x1<100 or x3>20'
        tk.Label(expressionFrame, text=helpText).grid(row=1, column=1, sticky='nw')

        # Buttons
        buttonFrame = tk.Frame(frame)
        buttonFrame.grid(row=5, column=1, sticky='sew', pady=(10, 5))
        buttonFrame.columnconfigure(0, weight=1)
        resetBtn = Button(buttonFrame, text='Reset', command=self._onResetClick)
        resetBtn.grid(row=0, column=0, sticky='ne', padx=(5, 0))
        updateBtn = Button(buttonFrame, text='Update Plot', imagePath='fa-refresh.png',
                           command=self._onUpdateClick)
        updateBtn.grid(row=0, column=1, sticky='ne', padx=5)

        frame.grid(row=1, column=0, sticky='new', padx=5, pady=(10, 5))

    def _onUpdateClick(self,e=None):
        if self.selectTool.get() == TOOL_TRAJECTORY :
            TrajectoriesWindow._onUpdateClick(self,e)
            self.createClusterBtn.config(state=tk.DISABLED)
            if (self.pathData.getSize() < self.numberOfPoints):
                self.updateClusterBtn.config(state=tk.NORMAL)
            self.eraseBtn.config(state=tk.DISABLED)
            self.trajSimBtn.config(state=tk.NORMAL)


        if self.selectTool.get() == TOOL_CLUSTERING:
            ClusteringWindow._onUpdateClick(self, e)
            self.createClusterBtn.config(state=tk.NORMAL)
            self.updateClusterBtn.config(state=tk.DISABLED)
            self.trajSimBtn.config(state=tk.DISABLED)
            self.eraseBtn.config(state=tk.NORMAL)

    def _exportBox(self,content):
        frame = tk.LabelFrame(content, text='Import/Export', font=self.fontBold)

        nameFrame = tk.Frame(frame)
        nameFrame.grid(row=0, column=0, sticky='w', pady=(10, 5))

        label = tk.Label(nameFrame, text="Name", font=self.fontBold)
        label.grid(row=0, column=0, padx=5, pady=5, sticky='w')
        self.clusterName = tk.StringVar()
        clusterEntry = tk.Entry(nameFrame, textvariable=self.clusterName,
                                width=30, bg='white')
        clusterEntry.grid(row=0, column=2, pady=5)


        buttonFrame = tk.Frame(frame)
        buttonFrame.grid(row=1, column=0, sticky='w', pady=(10, 5))

        self.saveClusterBtn = Button(buttonFrame, text='Export to EM dataset', state=tk.DISABLED,
                              tooltip='export clusters to scipion', command=self._onSaveClusterClick)
        self.saveClusterBtn.grid(row=0, column=2, padx=5)

        self.saveBtn = Button(buttonFrame, text='Save animation state',
                              tooltip='Save the trajectory', command=self._onSaveClick)
        self.saveBtn.grid(row=0, column=3)

        self.loadBtn = Button(buttonFrame, text='Load animation state', imagePath='fa-folder-open.png',
                              tooltip='Load a previous PCA clustering', command=self._onLoadClick)
        self.loadBtn.grid(row=0, column=4)

        frame.grid(row=4, column=0, sticky='new', padx=5, pady=(10, 5))



    def _createClusteringBox(self, content):
        frame = tk.LabelFrame(content, text='Clustering', font=self.fontBold)
        frame.columnconfigure(0, minsize=50)
        frame.columnconfigure(1, weight=1)


        buttonsFrame = tk.Frame(frame)
        buttonsFrame.grid(row=0, column=0,
                          sticky='new', padx=5, pady=5)

        label = tk.Label(buttonsFrame, text="From trajectory", font = self.fontItalic)
        label.grid(row=0, column=0, padx=5, pady=5, sticky='w')

        self.updateClusterBtn = Button(buttonsFrame, text='Cluster from traj', state=tk.DISABLED,
                                     tooltip='Create new cluster',
                                     imagePath='fa-plus-circle.png', command=self._onUpdateCluster)
        self.updateClusterBtn.grid(row=0, column=1, padx=5)


        buttonsFrame = tk.Frame(frame)
        buttonsFrame.grid(row=1, column=0,
                          sticky='new', padx=5, pady=5)

        label = tk.Label(buttonsFrame, text="From selection", font = self.fontItalic)
        label.grid(row=0, column=0, padx=5, pady=5, sticky='w')

        self.createClusterBtn = Button(buttonsFrame, text='New cluster from sel', state=tk.DISABLED,
                                     tooltip='New clutser from sel',
                                     imagePath='fa-plus-circle.png', command=self._onCreateCluster)
        self.createClusterBtn.grid(row=0, column=1, padx=5)

        self.eraseBtn = Button(buttonsFrame, text='Erase sel',  tooltip='Erase selection', command=self._onErase)
        self.eraseBtn.grid(row=0, column=2, padx=5)

        buttonsFrame = tk.Frame(frame)
        buttonsFrame.grid(row=2, column=0,
                          sticky='new', padx=5, pady=5)

        label = tk.Label(buttonsFrame, text="K-means", font = self.fontItalic)
        label.grid(row=0, column=0, padx=5, pady=5, sticky='w')

        kmeansCluster = Button(buttonsFrame, text='Compute clusters', state=tk.NORMAL,
                                     tooltip='K means clustering', command=self._onKMeansCluster)
        kmeansCluster.grid(row=0, column=1, padx=5)

        label = tk.Label(buttonsFrame, text="Number of clusters")
        label.grid(row=0, column=2, padx=5, pady=5, sticky='w')
        self.numPointsKmean = tk.StringVar(value="3")
        clusterEntry = tk.Entry(buttonsFrame, textvariable=self.numPointsKmean ,
                                width=3, bg='white')
        clusterEntry.grid(row=0, column=3, pady=5)


        frame.grid(row=3, column=0, sticky='new', padx=5, pady=(10, 5))

    def _createTrajectoriesBox(self, content):
        frame = tk.LabelFrame(content, text='Trajectories', font=self.fontBold, highlightcolor="cyan")
        buttonsFrame = tk.Frame(frame)
        buttonsFrame.grid(row=0, column=0,
                          sticky='w', padx=5, pady=5)

        label = tk.Label(buttonsFrame, text="Number of points")
        label.grid(row=0, column=0, padx=5, pady=5, sticky='w')
        self.numberOfPointsVar = tk.StringVar(value=str(self.numberOfPoints))
        nPointsEntry = tk.Entry(buttonsFrame, textvariable=self.numberOfPointsVar ,
                                width=3, bg='white')
        nPointsEntry.grid(row=0, column=1, pady=5)
        setNPointsBtn = Button(buttonsFrame, text='Set', state=tk.NORMAL,
                                tooltip='Set the number of points for the trajectory', command=self._onSetNPoints)
        setNPointsBtn.grid(row=0, column=2, padx=5)


        buttonsFrame2 = tk.Frame(frame)
        buttonsFrame2.grid(row=1, column=0,
                          sticky='w', padx=5, pady=5)
        buttonsFrame2.columnconfigure(0, weight=1)
        self.trajSimBtn = Button(buttonsFrame2, text='Generate points', state=tk.NORMAL,
                                     tooltip='Generate trajectory points based on axis and trajectory type', command=self._onSimClick)
        self.trajSimBtn.grid(row=0, column=0, padx=5)
        self.trajAxisBtn = ComboBox(buttonsFrame2, choices=["axis %i"%(i+1) for i in range(self.dim)])
        self.trajAxisBtn.grid(row=0, column=1, padx=(5, 10))
        self.trajTypeBtn = ComboBox(buttonsFrame2, choices=["percentiles",
                                                            "linear betmeen min and max", "Linear betmeen -2*std and +2*std"
                                                            , "Gaussian betmeen min and max", "Gaussian betmeen -2*std and +2*std"])
        self.trajTypeBtn.grid(row=0, column=2, padx=(5, 5))

        buttonsFrame3 = tk.Frame(frame)
        buttonsFrame3.grid(row=2, column=0,
                          sticky='w', padx=5, pady=5)
        buttonsFrame3.columnconfigure(0, weight=1)
        self.generateBtn = Button(buttonsFrame3, text='Show in VMD', state=tk.NORMAL,
                                     tooltip='Select trajectory points to generate the animations',
                                     imagePath='fa-plus-circle.png', command=self._onCreateClick)
        self.generateBtn.grid(row=0, column=0, padx=5)
        self.comboBtn = ComboBox(buttonsFrame3, choices=["Inverse transformation", "cluster average"])
        self.comboBtn.grid(row=0, column=1, padx=(5, 10))

        frame.grid(row=2, column=0, sticky='new', padx=5, pady=(5, 10))

    def _onSaveClusterClick(self, e=None):
        if self.saveClusterCallback:
            self.saveClusterCallback(self)

    def _onSaveClick(self, e=None):
        if self.saveCallback:
            self.saveCallback(self)

    def _onSetNPoints(self):
        try :
            self.numberOfPoints = int(self.numberOfPointsVar.get())
        except:
            self.showError("Can not read number of points.")

    def _onKMeansCluster(self):
        try :
            n_clusters = int(self.numPointsKmean.get())
        except:
            return self.showError("Can not read number of clusters")
        self._onUpdateClick()

        k_means = KMeans(init='k-means++', n_clusters=n_clusters)
        selection = np.array(self.listbox.curselection())
        data_arr = np.array([p.getData()[selection] for p in self.data])
        k_means.fit(data_arr)

        classes = k_means.labels_ + 1
        i=0
        for point in self.data:
            point._weight = classes[i]
            i+=1
        self.saveClusterBtn.config(state=tk.NORMAL)
        self._onUpdateClick()
        self.setClusterNumber(3)

    def _onSimClick(self, e=None):
        TrajectoriesWindow._onResetClick(self, e)
        traj_axis = self.trajAxisBtn.getValue()
        traj_type = self.trajTypeBtn.getValue()

        data_axis = np.array([p.getData()[traj_axis] for p in self.data])
        mean_axis =data_axis.mean()
        std_axis =data_axis.std()
        min_axis =data_axis.min()
        max_axis =data_axis.max()

        traj_points = np.zeros((self.numberOfPoints, self.dim))
        if traj_type== 0 :
            traj_points[:,traj_axis] = np.array(
                [np.percentile(data_axis,100*(i+1)/(self.numberOfPoints+1)) for i in range(self.numberOfPoints)]
            )
        elif traj_type== 1 :
            traj_points[:,traj_axis] = np.linspace(min_axis,max_axis,self.numberOfPoints)
        elif traj_type== 2:
            traj_points[:,traj_axis] = np.linspace(-2*std_axis,+2*std_axis,self.numberOfPoints)
        elif traj_type== 3:
            distribution = sp.stats.norm(loc=mean_axis, scale=std_axis)
            bounds_for_range = distribution.cdf([min_axis, max_axis])
            gaussTraj = distribution.ppf(np.linspace(*bounds_for_range, num=self.numberOfPoints))
            traj_points[:, traj_axis] = gaussTraj
        elif traj_type== 4:
            distribution = sp.stats.norm(loc=mean_axis, scale=std_axis)
            bounds_for_range = distribution.cdf([-2*std_axis, +2*std_axis])
            gaussTraj = distribution.ppf(np.linspace(*bounds_for_range, num=self.numberOfPoints))
            traj_points[:, traj_axis] = gaussTraj

        for i in range(self.numberOfPoints):
            self.pathData.addPoint(Point(pointId=i + 1, data=traj_points[i], weight=0))

        self._checkNumberOfPoints()
        self._onUpdateClick()

    def _onUpdateCluster(self):
        traj_arr = np.array([p.getData() for p in self.pathData])
        selection = np.array(self.listbox.curselection())
        traj_sel = traj_arr[:, selection]

        for point in self.data:
            point_sel = point.getData()[selection]
            closet_point = np.argmin(np.linalg.norm(traj_sel - point_sel, axis=1))
            point._weight = closet_point + 1

        self.saveClusterBtn.config(state=tk.NORMAL)
        self._onUpdateClick()
        self.setClusterNumber(self.numberOfPoints)


    def _onCreateCluster(self):
        self.setClusterNumber(self.getClusterNumber() +1)
        for point in self.data:
            if point.getState() == Point.SELECTED:
                point._weight =self.getClusterNumber()
        self.saveClusterBtn.config(state=tk.NORMAL)
        ClusteringWindow._onResetClick(self)

    def setClusterNumber(self, n):
        self._clusterNumber = n

    def _onErase(self):
        for point in self.data:
            if point.getState() == Point.SELECTED:
                point._weight =0.0
        self.saveClusterBtn.config(state=tk.NORMAL)
        ClusteringWindow._onResetClick(self)

    def _checkNumberOfPoints(self):
        TrajectoriesWindow._checkNumberOfPoints(self)
        self.updateClusterBtn.config(state=tk.NORMAL)

    def _onResetClick(self, e=None):
        self.updateClusterBtn.config(state=tk.DISABLED)
        self.saveClusterBtn.config(state=tk.DISABLED)
        self.setClusterNumber(0)

        for point in self.data:
            point._weight = 0
        TrajectoriesWindow._onResetClick(self, e)
        self.generateBtn.config(state=tk.NORMAL)


    def getClusterName(self):
        return self.clusterName.get().strip()

    def getAnimationType(self):
        return self.comboBtn.getValue()

    def getClusterNumber(self):
        return self._clusterNumber


# **************************************************************************
# * Author:  Remi Vuillemot
# * IMPMC, UPMC Sorbonne University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# **************************************************************************

import joblib
from pyworkflow.protocol.params import (PointerParam, EnumParam, IntParam)
from pwem.protocols import ProtAnalysis3D
from pyworkflow.utils.path import makePath, copyFile
from pyworkflow.protocol import params
from pwem.emlib import MetaData, MDL_ENABLED, MDL_NMA_MODEFILE,MDL_ORDER
from pwem.objects import SetOfNormalModes, AtomStruct
from .convert import rowToMode
from xmipp3.base import XmippMdRow
from continuousflex.protocols.utilities.genesis_utilities import numpyArr2dcd, dcd2numpyArr

import numpy as np
import glob
from sklearn import decomposition
from joblib import dump

from .utilities.genesis_utilities import dcd2numpyArr
from .utilities.pdb_handler import ContinuousFlexPDBHandler
import pwem.emlib.metadata as md
import continuousflex
from continuousflex import Plugin
from subprocess import check_call
import sys

PDB_SOURCE_SUBTOMO = 0
PDB_SOURCE_PATTERN = 1
PDB_SOURCE_OBJECT = 2
PDB_SOURCE_TRAJECT = 3
PDB_SOURCE_ALIGNED = 4

REDUCE_METHOD_PCA = 0
REDUCE_METHOD_UMAP = 1


class FlexProtDimredPdb(ProtAnalysis3D):
    """ Protocol for applying dimentionality reduction on PDB files. """
    _label = 'pdb dimentionality reduction'


    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('pdbSource', EnumParam, default=PDB_SOURCE_SUBTOMO,
                      label='Source of PDBs',
                      choices=['Used for subtomogram synthesis', 'File pattern', 'Object', 'Trajectory Files', 'Align PDBs protocol'],
                      help='Use the file pattern as file location with /*.pdb')
        form.addParam('pdbs', params.PointerParam, pointerClass='FlexProtSynthesizeSubtomo',
                      condition='pdbSource ==  %i'%PDB_SOURCE_SUBTOMO,
                      label="Subtomogram synthesis",
                      help='Point to a protocol of synthesizing subtomograms, the ground truth PDBs will be used as input')
        form.addParam('pdbs_file', params.PathParam,
                      condition='pdbSource == %i' % PDB_SOURCE_PATTERN,
                      label="List of PDBs",
                      help='Use the file pattern as file location with /*.pdb')
        form.addParam('setOfPDBs', params.PointerParam, pointerClass='SetOfPDBs, SetOfAtomStructs',
                      condition='pdbSource == %i' % PDB_SOURCE_OBJECT,
                      label="Set of PDBs",
                      help='Use a scipion object SetOfPDBs / SetOfAtomStructs')
        form.addParam('dcds_file', params.PathParam,
                      condition='pdbSource == %i' % PDB_SOURCE_TRAJECT,
                      label="DCD trajectory file (s)",
                      help='Use the file pattern as file location with /*.dcd')
        form.addParam('dcd_ref_pdb', params.PointerParam, pointerClass='AtomStruct',
                      condition='pdbSource == %i' % PDB_SOURCE_TRAJECT,
                      label="trajectory Reference PDB",
                      help='Reference PDB of the trajectory (Only used for structural information (Atom name, residue number etc)'
                           '. The coordinates inside this PDB are not used. The atoms number and position in the file must'
                           ' correspond to the DCD file. ')
        form.addParam('dcd_start', params.IntParam, default=0,
                      condition='pdbSource == %i' % PDB_SOURCE_TRAJECT,
                      label="Beginning of the trajectory",
                      help='Index of the desired begining of the trajectory', expertLevel=params.LEVEL_ADVANCED)
        form.addParam('dcd_end', params.IntParam, default=-1,
                      condition='pdbSource == %i' % PDB_SOURCE_TRAJECT,
                      label="Ending of the trajectory",
                      help='Index of the desired end of the trajectory', expertLevel=params.LEVEL_ADVANCED)
        form.addParam('dcd_step', params.IntParam, default=1,
                      condition='pdbSource == %i' % PDB_SOURCE_TRAJECT,
                      label="Step of the trajectory",
                      help='Step to skip points in the trajectory', expertLevel=params.LEVEL_ADVANCED)

        form.addParam('alignPdbProt', params.PointerParam, pointerClass='FlexProtAlignPdb',
                      condition='pdbSource == %i' % PDB_SOURCE_ALIGNED,
                      label="Align PDBs Protocol",
                      help='Point to a protocol of pdb aligned. For large data set, you can use here the align pdb protocol as input '
                           'and avoid creating an output set of pdb in the align pdb protocol.')

        form.addParam('method', params.EnumParam, label="Reduction method", default=REDUCE_METHOD_PCA,
                      choices=['PCA', 'UMAP'],help="")

        form.addParam('reducedDim', IntParam, default=10,
                      label='Number of Principal Components')

        # --------------------------- INSERT steps functions --------------------------------------------
    def _insertAllSteps(self):
        self._insertFunctionStep('readInputFiles')
        self._insertFunctionStep('performDimred')
        if self.method.get() == REDUCE_METHOD_PCA:
            self._insertFunctionStep('createOutputStep')

    # --------------------------- STEPS functions --------------------------------------------
    def readInputFiles(self):
        inputFiles = self.getInputFiles()

        # Get pdbs coordinates
        if self.pdbSource.get() == PDB_SOURCE_TRAJECT:
            pdbs_arr = dcd2numpyArr(inputFiles[0])
            for i in range(1,len(inputFiles)):
                pdb_arr_i = dcd2numpyArr(inputFiles[i])
                pdbs_arr = np.concatenate((pdbs_arr, pdb_arr_i), axis=0)

        elif self.pdbSource.get() == PDB_SOURCE_ALIGNED:
            pdbs_arr = dcd2numpyArr(inputFiles[0])
        else:
            pdbs_matrix = []
            for pdbfn in inputFiles:
                try:
                    # Read PDBs
                    mol = ContinuousFlexPDBHandler(pdbfn)
                    pdbs_matrix.append(mol.coords)
                except RuntimeError:
                    print("Warning : Can not read PDB file %s " % pdbfn)
            pdbs_arr = np.array(pdbs_matrix)

        # save as dcd file
        numpyArr2dcd(pdbs_arr, self._getExtraPath("coords.dcd"))

    def performDimred(self):

        pdbs_arr = dcd2numpyArr(self._getExtraPath("coords.dcd"))
        nframe, natom,_ = pdbs_arr.shape
        pdbs_matrix = pdbs_arr.reshape(nframe, natom*3)

        if self.method.get() == REDUCE_METHOD_PCA:
            pca = decomposition.PCA(n_components=self.reducedDim.get())
            Y = pca.fit_transform(pdbs_matrix)
            dump(pca, self._getExtraPath('pca_pickled.joblib'))

            pathPC = self._getPath("modes")
            pdb = ContinuousFlexPDBHandler(self.getPDBRef())
            pdb.coords = pca.mean_.reshape(pdbs_matrix.shape[1] // 3, 3)
            pdb.write_pdb(self._getPath("atoms.pdb"))
            makePath(pathPC)
            matrix = pca.components_.reshape(self.reducedDim.get(),pdbs_matrix.shape[1]//3,3)
            self.writePrincipalComponents(prefix=pathPC, matrix = matrix)

        elif self.method.get() == REDUCE_METHOD_UMAP:
            pdbs_dump = self._getTmpPath('pdbs_dump.pkl')
            joblib.dump(pdbs_matrix, pdbs_dump)
            Y_dump = self._getTmpPath('Y_dump.pkl')
            args = "%d %d %d %s %s %s" % (self.reducedDim.get(), 15, 1000,
                                       pdbs_dump, self._getExtraPath('pca_pickled.joblib'), Y_dump)
            script_path = continuousflex.__path__[0] + '/protocols/utilities/umap_run.py '
            command = "python " + script_path + args
            command = Plugin.getContinuousFlexCmd(command)
            check_call(command, shell=True, stdout=sys.stdout, stderr=sys.stderr,
                       env=None, cwd=None)
            Y = joblib.load(Y_dump)

        np.savetxt(self.getOutputMatrixFile(),Y)

    def createOutputStep(self):
            # Metadata
            mdOut = MetaData()
            for i in range(self.reducedDim.get()):
                objId = mdOut.addObject()
                modefile = self._getPath("modes", "vec.%d" % (i + 1))
                mdOut.setValue(MDL_NMA_MODEFILE, modefile, objId)
                mdOut.setValue(MDL_ORDER, i + 1, objId)
                mdOut.setValue(MDL_ENABLED, 1, objId)
            mdOut.write(self._getPath("modes.xmd"))

            # Sqlite object
            pcSet =SetOfNormalModes(filename=self._getPath("modes.sqlite"))
            row = XmippMdRow()
            for objId in mdOut:
                row.readFromMd(mdOut, objId)
                pcSet.append(rowToMode(row))

            pdb = AtomStruct(self._getPath("atoms.pdb"))
            self._defineOutputs(outputMean=pdb)

            pcSet.setPdb(pdb)
            self._defineOutputs(outputPCA=pcSet)

    # --------------------------- INFO functions --------------------------------------------
    def _summary(self):
        summary = []
        return summary

    def _validate(self):
        errors = []
        return errors

    def _citations(self):
        return ['harastani2022continuousflex']

    def _methods(self):
        pass

    # --------------------------- UTILS functions --------------------------------------------
    def _printWarnings(self, *lines):
        """ Print some warning lines to 'warnings.xmd',
        the function should be called inside the working dir."""
        fWarn = open("warnings.xmd", 'w')
        for l in lines:
            print >> fWarn, l
        fWarn.close()

    def getInputFiles(self):
        if self.pdbSource.get()==PDB_SOURCE_SUBTOMO:
            l= [f for f in glob.glob(self.pdbs.get()._getExtraPath('*.pdb'))]
        elif self.pdbSource.get()==PDB_SOURCE_PATTERN:
            l= [f for f in glob.glob(self.pdbs_file.get())]
        elif self.pdbSource.get()==PDB_SOURCE_OBJECT:
            l= [i.getFileName() for i in self.setOfPDBs.get()]
        elif self.pdbSource.get()==PDB_SOURCE_TRAJECT:
            l= [f for f in glob.glob(self.dcds_file.get())]
        elif self.pdbSource.get()==PDB_SOURCE_ALIGNED:
            l=[self.alignPdbProt.get()._getExtraPath("coords.dcd")]
        l.sort()
        return l

    def getPDBRef(self):
        if self.pdbSource.get()==PDB_SOURCE_TRAJECT:
            return self.dcd_ref_pdb.get().getFileName()
        elif self.pdbSource.get()==PDB_SOURCE_ALIGNED:
            return self.alignPdbProt.get()._getExtraPath("reference.pdb")
        else:
            return self.getInputFiles()[0]

    def getOutputMatrixFile(self):
        return self._getExtraPath('output_matrix.txt')

    def writePrincipalComponents(self, prefix, matrix):
        for i in range(self.reducedDim.get()):
            with open("%s/vec.%i"%(prefix,i+1), "w") as f:
                    for j in range(matrix.shape[1]):
                        f.write(" %e   %e   %e\n" % (matrix[i,j, 0], matrix[i,j, 1], matrix[i,j, 1]))



# **************************************************************************
# * Authors:    Mohamad Harastani            (mohamad.harastani@igbmc.fr)
# *             Slavica Jonic                (slavica.jonic@upmc.fr)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from os.path import basename
from pyworkflow.viewer import (ProtocolViewer, DESKTOP_TKINTER, WEB_DJANGO)
from pyworkflow.protocol.params import StringParam, LEVEL_ADVANCED
from pyworkflow.protocol import params
from continuousflex.protocols.protocol_nma_alignment_vol import FlexProtAlignmentNMAVol
from continuousflex.protocols.protocol_subtomograms_synthesize import FlexProtSynthesizeSubtomo
from continuousflex.protocols.data import Point, Data
from pwem.emlib import MetaData, MDL_ORDER, MDL_ANGLE_ROT, MDL_ANGLE_TILT, MDL_ANGLE_PSI, MDL_SHIFT_X, MDL_SHIFT_Y, \
    MDL_SHIFT_Z, MDL_NMA
from .plotter_vol import FlexNmaVolPlotter
from continuousflex.protocols.convert import l2
from xmippLib import SymList
import numpy as np
import tkinter.messagebox as mb
import matplotlib.pyplot as plt

FIGURE_LIMIT_NONE = 0
FIGURE_LIMITS = 1

X_LIMITS_NONE = 0
X_LIMITS = 1
Y_LIMITS_NONE = 0
Y_LIMITS = 1
Z_LIMITS_NONE = 0
Z_LIMITS = 1

METADATA_PROJECT = 0
METADATA_FILE = 1


class FlexAlignmentNMAVolViewer(ProtocolViewer):
    """ Visualization of results from the NMA alignment vol protocol (HEMNMA-3D)
    """
    _label = 'viewer nma alignment vol'
    _targets = [FlexProtAlignmentNMAVol]
    _environments = [DESKTOP_TKINTER, WEB_DJANGO]

    def __init__(self, **kwargs):
        ProtocolViewer.__init__(self, **kwargs)
        self._data = None

    def getData(self):
        if self._data is None:
            self._data = self.loadData()
        return self._data

    def _defineParams(self, form):
        form.addSection(label='Visualization')
        form.addParam('displayRawDeformation', StringParam, default='7 8',
                      label='Display raw deformation',
                      help='Type 7 to see the histogram of raw deformation along mode 7; \n'
                           'type 8 to see the histogram of raw deformation along mode 8, etc.\n'
                           'Type 7 8 to see the 2D plot of raw deformations along modes 7 vs 8.\n'
                           'Type 7 8 9 to see the 3D plot of raw deformations along modes 7, 8 and 9; etc.'
                      )
        form.addParam('limits_modes', params.EnumParam,
                      choices=['Automatic (Recommended)', 'Set manually Use upper and lower values'],
                      default=FIGURE_LIMIT_NONE,
                      label='(1 - CC) limits', display=params.EnumParam.DISPLAY_COMBO,
                      help='If you want to use a range of (1-CC) choose to set it manually.')
        form.addParam('LimitLow', params.FloatParam, default=None,
                      condition='limits_modes==%d' % FIGURE_LIMITS,
                      label='Lower (1-CC) value',
                      help='The lower (1-CC) used in the graph')
        form.addParam('LimitHigh', params.FloatParam, default=None,
                      condition='limits_modes==%d' % FIGURE_LIMITS,
                      label='Upper (1-CC) value',
                      help='The upper (1-CC) used in the graph')
        form.addParam('xlimits_mode', params.EnumParam,
                      choices=['Automatic (Recommended)', 'Set manually x-axis limits'],
                      default=X_LIMITS_NONE,
                      label='x-axis limits', display=params.EnumParam.DISPLAY_COMBO,
                      help='This allows you to use a specific range of x-axis limits')
        form.addParam('xlim_low', params.FloatParam, default=None,
                      condition='xlimits_mode==%d' % X_LIMITS,
                      label='Lower x-axis limit')
        form.addParam('xlim_high', params.FloatParam, default=None,
                      condition='xlimits_mode==%d' % X_LIMITS,
                      label='Upper x-axis limit')
        form.addParam('ylimits_mode', params.EnumParam,
                      choices=['Automatic (Recommended)', 'Set manually y-axis limits'],
                      default=Y_LIMITS_NONE,
                      label='y-axis limits', display=params.EnumParam.DISPLAY_COMBO,
                      help='This allows you to use a specific range of y-axis limits')
        form.addParam('ylim_low', params.FloatParam, default=None,
                      condition='ylimits_mode==%d' % Y_LIMITS,
                      label='Lower y-axis limit')
        form.addParam('ylim_high', params.FloatParam, default=None,
                      condition='ylimits_mode==%d' % Y_LIMITS,
                      label='Upper y-axis limit')
        form.addParam('zlimits_mode', params.EnumParam,
                      choices=['Automatic (Recommended)', 'Set manually z-axis limits'],
                      default=Z_LIMITS_NONE,
                      label='z-axis limits', display=params.EnumParam.DISPLAY_COMBO,
                      help='This allows you to use a specific range of z-axis limits')
        form.addParam('zlim_low', params.FloatParam, default=None,
                      condition='zlimits_mode==%d' % Z_LIMITS,
                      label='Lower z-axis limit')
        form.addParam('zlim_high', params.FloatParam, default=None,
                      condition='zlimits_mode==%d' % Z_LIMITS,
                      label='Upper z-axis limit')
        group = form.addGroup('Comparing with ground-truth', expertLevel=LEVEL_ADVANCED)
        group.addParam('GroundTruth', params.EnumParam,
                       choices=['From volume synthesis protocol', 'From an external metadata file'],
                       default=METADATA_PROJECT,
                       label='Ground-Truth parameters', display=params.EnumParam.DISPLAY_COMBO,
                       help='Use this is only when testing the method with synthetic data')
        group.addParam('SynthesisProject', params.PointerParam, pointerClass='FlexProtSynthesizeSubtomo',
                       condition='GroundTruth==%d' % METADATA_PROJECT,
                       allowsNull=True,
                       label="Project for volume synthesize",
                       help='Select a previous run for volume (subtomograms) synthesize.')
        group.addParam('MetadataFile', params.FileParam,
                       pointerClass='params.FileParam', allowsNull=True,
                       condition='GroundTruth==%d' % METADATA_FILE,
                       label="Metadata file (xmd)",
                       help='Choose a metadata file containing angles, shifts and NM amplitudes, typically a metadata'
                            ' file from synthesizing volumes')
        group.addParam('displayStatistics', params.LabelParam,
                       label="Display error statistics and plots?")

    def _getVisualizeDict(self):
        return {'displayRawDeformation': self._viewRawDeformation,
                'displayStatistics': self._viewErrorStatistics,
                }


    def _viewRawDeformation(self, paramName):
        components = self.displayRawDeformation.get()
        return self._doViewRawDeformation(components)

    def _doViewRawDeformation(self, components):
        components = list(map(int, components.split()))
        dim = len(components)
        views = []

        if dim > 0:
            modeList = []
            modeNameList = []
            missingList = []

            for modeNumber in components:
                found = False
                md = MetaData(self.protocol._getExtraPath('modes.xmd'))

                for i, objId in enumerate(md):
                    modeId = md.getValue(MDL_ORDER, objId)

                    if modeNumber == modeId:
                        modeNameList.append('Mode %d' % modeNumber)
                        modeList.append(i)
                        found = True
                        break
                if not found:
                    missingList.append(str(modeNumber))

            if missingList:
                return [self.errorMessage("Invalid mode(s) *%s*\n." % (', '.join(missingList)),
                                          title="Invalid input")]

            # Actually plot
            if self.limits_modes == FIGURE_LIMIT_NONE:
                plotter = FlexNmaVolPlotter(data=self.getData(),
                                            xlim_low=self.xlim_low, xlim_high=self.xlim_high,
                                            ylim_low=self.ylim_low, ylim_high=self.ylim_high,
                                            zlim_low=self.zlim_low, zlim_high=self.zlim_high)
            else:
                plotter = FlexNmaVolPlotter(data=self.getData(),
                                            LimitL=self.LimitLow, LimitH=self.LimitHigh,
                                            xlim_low=self.xlim_low, xlim_high=self.xlim_high,
                                            ylim_low=self.ylim_low, ylim_high=self.ylim_high,
                                            zlim_low=self.zlim_low, zlim_high=self.zlim_high)

            baseList = [basename(n) for n in modeNameList]

            if dim == 1:
                self.getData().XIND = modeList[0]
                plotter.plotArray1D("Histogram for %s" % baseList[0],
                                    "Deformation Value", "Number of Volumes")
            else:
                self.getData().XIND = modeList[0]
                self.getData().YIND = modeList[1]
                if dim == 2:
                    plotter.plotArray2D("%s vs %s" % tuple(baseList), *baseList)
                elif dim == 3:
                    self.getData().ZIND = modeList[2]
                    plotter.plotArray3D("%s %s %s" % tuple(baseList), *baseList)
            views.append(plotter)

        return views

    def _viewErrorStatistics(self, paramName):
        if self.GroundTruth.get() == METADATA_PROJECT:
            metadata_file = self.SynthesisProject.get()._getExtraPath('GroundTruth.xmd')
        else:
            metadata_file = self.MetadataFile.get()
        return self._doViewErrorStatistics(metadata_file)

    def _doViewErrorStatistics(self, metadata_file):
        md_gt = MetaData(metadata_file)
        md_protocol = MetaData(self.protocol._getExtraPath('volumes.xmd'))
        md_protocol.sort()
        # Get the matching modes:
        md_modes = MetaData(self.protocol._getExtraPath('modes.xmd'))
        modeIds = []
        for i, objId in enumerate(md_modes):
            modeIds.append(md_modes.getValue(MDL_ORDER, objId))
        # Get the parameters from both lists:
        rtp_protocol = []
        xyz_protocol = []
        mode_ampl_protocol = []
        rtp_gt = []
        xyz_gt = []
        mode_ampl_gt = []
        for objId in md_protocol:
            rtp_protocol.append([md_protocol.getValue(MDL_ANGLE_ROT, objId),
                                 md_protocol.getValue(MDL_ANGLE_TILT, objId),
                                 md_protocol.getValue(MDL_ANGLE_PSI, objId)])
            xyz_protocol.append([md_protocol.getValue(MDL_SHIFT_X, objId),
                                 md_protocol.getValue(MDL_SHIFT_Y, objId),
                                 md_protocol.getValue(MDL_SHIFT_Z, objId)])
            mode_ampl_protocol.append(md_protocol.getValue(MDL_NMA, objId))

            rtp_gt.append([md_gt.getValue(MDL_ANGLE_ROT, objId),
                           md_gt.getValue(MDL_ANGLE_TILT, objId),
                           md_gt.getValue(MDL_ANGLE_PSI, objId)])
            xyz_gt.append([md_gt.getValue(MDL_SHIFT_X, objId),
                           md_gt.getValue(MDL_SHIFT_Y, objId),
                           md_gt.getValue(MDL_SHIFT_Z, objId)])
            mode_ampl_gt.append(md_gt.getValue(MDL_NMA, objId))

        # Angular and shift distances
        shift_distance = []
        angular_distance = []
        # The full description of computeDistanceAngles function is:
        # A = SymList.computeDistanceAngles(SymList(), rot1, tilt1, psi1, rot2, tilt2, psi2, projdir_mode, check_mirrors, object_rotation)
        # By default, they are all set to False. However, check_mirrors should be true in general.
        for i in range(len(rtp_protocol)):
            shift_distance.append(l2(xyz_gt[i], xyz_protocol[i]))
            angular_distance.append(SymList.computeDistanceAngles(SymList(),
                                                                  rtp_protocol[i][0], rtp_protocol[i][1], rtp_protocol[i][2],
                                                                  rtp_gt[i][0], rtp_gt[i][1], rtp_gt[i][2],
                                                                  False, True, False))
        # Normal mode amplitudes distances: we need to find the subset of normal modes used in alignment in the groundtruth
        counter = 0
        plt.figure()
        mean_amplitudes = []
        std_amplitudes = []
        label = []
        dist = []
        for i in modeIds:
            # mode 7 corresponds to zero in the ground truth, so we need to subtract 7
            A = np.array(mode_ampl_gt)[:,i - 7]
            B = np.array(mode_ampl_protocol)[:, counter]
            mean_amplitudes.append(np.mean(np.array(A - B)))
            std_amplitudes.append(np.std(np.array(A - B)))
            label.append('mode ' + str(i))
            dist.append(np.array(A - B))
            counter +=1
        plt.title('histogram of normal mode amplitude distances')
        plt.hist(dist, bins=100, label=label)
        plt.legend(loc='upper right')

        plt.figure()
        plt.hist(np.array(angular_distance), bins=100)
        plt.title('histogram of angular distance')
        plt.figure()
        plt.hist(np.array(shift_distance), bins=100)
        plt.title('histogram of shift distance')

        message = 'mean and standard deviation angular distance: ' + str(np.mean(np.array(angular_distance)))[:7]
        message += ' and ' + str(np.std(np.array(angular_distance)))[:7]
        message += '\nmean and standard deviation shift distance: ' + str(np.mean(np.array(shift_distance)))[:7]
        message += ' and ' + str(np.std(np.array(shift_distance)))[:7]

        counter = 0
        for i in modeIds:
            message += '\nmean and standard deviation for mode ' + str(i) + ': ' + str(mean_amplitudes[counter])[:7] + \
                       ' ' + str(std_amplitudes[counter])[:7]
            counter +=1

        mb.showinfo('Distances compared to the ground truth', message)
        plt.show()
        pass

    def loadData(self):
        """ Iterate over the images and their deformations 
        to create a Data object with theirs Points.
        """
        particles = self.protocol.outputParticles
        data = Data()
        for i, particle in enumerate(particles):
            pointData = list(map(float, particle._xmipp_nmaDisplacements))
            data.addPoint(Point(pointId=particle.getObjId(),
                                data=pointData,
                                weight=particle._xmipp_maxCC.get()))
        return data



# * Authors:  Mohamad Harastani          (mohamad.harastani@igbmc.fr)
# *           Rémi Vuillemot             (remi.vuillemot@upmc.fr)
# *
# * IMPMC, UPMC Sorbonne University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# **************************************************************************

import os
import time

import pwem.emlib.metadata as md
import pyworkflow.protocol.params as params
import xmipp3
from pwem.protocols import ProtAnalysis3D
from pwem.utils import runProgram
from pyworkflow.protocol.params import NumericRangeParam
from pyworkflow.utils import getListFromRangeString
from pyworkflow.utils import replaceExt, getExt, replaceBaseExt
from pwem.convert import cifToPdb
from xmipp3.convert import (writeSetOfParticles)
from pyworkflow.utils.path import copyFile, createLink
import numpy as np
import glob
from joblib import dump
from math import cos, sin, pi
from continuousflex.protocols.convert import matrix2eulerAngles

NMA_ALIGNMENT_WAV = 0
NMA_ALIGNMENT_PROJ = 1

MODE_RELATION_LINEAR = 0
MODE_RELATION_3CLUSTERS = 1
MODE_RELATION_MESH = 2
MODE_RELATION_RANDOM = 3
MODE_RELATION_PARABOLA = 4

LOWPASS_YES = 0
LOWPASS_NO = 1

ROTATION_SHIFT_YES = 0
ROTATION_SHIFT_NO = 1

RECONSTRUCTION_FOURIER = 0
RECONSTRUCTION_WBP = 1

NOISE_CTF_YES = 0
NOISE_CTF_NO = 1

NMA_NO = 0
NMA_YES = 1

ROTATION_UNIFORM = 0
ROTATION_GAUSS = 1


class FlexProtSynthesizeImages(ProtAnalysis3D):
    """ Protocol for synthesizing images. """
    _label = 'synthesize images'

    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('confVar', params.BooleanParam, default=NMA_YES,
                      label='Simulate conformational variability?',
                      help='If yes, you need to use normal mode analysis. If no, the given volume or atomic structure '
                           'will be projected into simulated particle images.')
        group1 = form.addGroup('Select an atomic structure OR an EM volume', condition='confVar==%d' % NMA_NO)
        group1.addParam('refAtomic', params.PointerParam ,pointerClass='AtomStruct', allowsNull=True, label='atomic structure')
        group1.addParam('refVolume', params.PointerParam ,pointerClass='Volume', allowsNull=True, label='EM volume')

        group2 = form.addGroup('conformational variability', condition='confVar==%d' % NMA_YES)
        group2.addParam('importPdbs', params.BooleanParam, default=False,
                        expertLevel=params.LEVEL_ADVANCED, label='Do you want to improt a set of heterogenous PDBs?',
                        help='if you keep this as No, you can use a set of normal modes to generate conformational '
                             'heterogeneity. Otherwise, set as Yes if you have a set of different PDBs that you want '
                             'to use to simulate subtomograms (for example a trajectory from molecular '
                             'dynamics simulation)')
        group2.addParam('pdbs_path', params.PathParam,
                      condition='importPdbs',
                      label="List of PDBs",
                      help='Use the file pattern as file location with /*.pdb')
        group2.addParam('inputModes', params.PointerParam, pointerClass='SetOfNormalModes',
                      allowsNull=True, condition='importPdbs==False',
                      label="Normal modes",
                      help='Set of modes computed by normal mode analysis.')
        group2.addParam('modeList', NumericRangeParam,
                      label="Modes selection", condition='importPdbs==False',
                      allowsNull=True,
                      default='7-8',
                      help='Select the normal modes that will be used for image synthesis. \n'
                           'It is usually two modes that should be selected, unless if the relationship is linear or random.\n'
                           'You have several ways to specify the modes.\n'
                           ' Examples:\n'
                           ' "7,8-10" -> [7,8,9,10]\n'
                           ' "8, 10, 12" -> [8,10,12]\n'
                           ' "8 9, 10-12" -> [8,9,10,11,12])\n')
        group2.addParam('modeRelationChoice', params.EnumParam, default=MODE_RELATION_LINEAR,
                      allowsNull=True, condition='importPdbs==False',
                      choices=['Linear relationship', '3 Clusters', 'Grid', 'Random', 'Upper half circle'],
                      label='Relationship between the modes',
                      help='linear relationship: all the selected modes will have equal amplitudes. \n'
                           '3 clusters: the volumes will be devided exactly into three classes.\n'
                           'Grid: the amplitudes will be in a grid shape (grid size is square of what grid step).\n'
                           'Random: all the amplitudes will be random in the given range.\n'
                           'Parabolic: The relationship between two modes will be upper half circle.')
        group2.addParam('centerPoint', params.IntParam, default=100,
                      allowsNull=True,
                      condition='importPdbs==False and modeRelationChoice==%d' % MODE_RELATION_3CLUSTERS,
                      label='Center point',
                      help='This number will be used to determine the distance between the clusters'
                           'center1 = (-center_point, 0)'
                           'center2 = (center_point, 0)'
                           'center3 = (0, center_point)')
        group2.addParam('modesAmplitudeRange', params.IntParam, default=150,
                        allowsNull=True,
                        condition='importPdbs==False and modeRelationChoice==%d or modeRelationChoice==%d or modeRelationChoice==%d'
                                  ' or modeRelationChoice==%d' % (MODE_RELATION_LINEAR, MODE_RELATION_MESH,
                                                                  MODE_RELATION_RANDOM, MODE_RELATION_PARABOLA),
                        label='Amplitude range N --> [-N, N]',
                        help='Choose the number N for which the generated normal mode amplitudes are in the range of [-N, N]')
        group2.addParam('meshRowPoints', params.IntParam, default=6,
                      allowsNull=True,
                      condition='importPdbs==False and modeRelationChoice==%d' % MODE_RELATION_MESH,
                      label='Grid number of steps',
                      help='This number will be the number of points in the row and the column (grid shape will be size*size)')
        form.addParam('numberOfVolumes', params.IntParam, default=36,
                      label='Number of images',
                      condition='importPdbs==False and modeRelationChoice!=%d'% MODE_RELATION_MESH,
                      help='Number of images that will be generated')
        form.addParam('samplingRate', params.FloatParam, default=2.2,
                      condition='confVar==%d or refAtomic!=None' % NMA_YES,
                      label='Sampling rate',
                      help='The sampling rate (pixel size in Angstroms)')
        form.addParam('volumeSize', params.IntParam, default=64,
                      condition='confVar==%d or refAtomic!=None' % NMA_YES,
                      label='Image size',
                      help='Image size in voxels (all volumes will be cubes)')
        form.addParam('seedOption', params.BooleanParam, default=True,
                      expertLevel=params.LEVEL_ADVANCED,
                      label='Random seed',
                      help='Keeping it as True means that different runs will generate different images in terms '
                           'of conforamtion and rigid-body parameters. If you set as False, then different runs will '
                           'have the same conformations and angles '
                           '(setting to False allows you to generate the same conformations and orientations with '
                           'different noise values).')

        form.addSection(label='Noise and CTF')
        form.addParam('noiseCTFChoice', params.EnumParam, default=ROTATION_SHIFT_YES,
                      choices=['Yes', 'No'],
                      label='Apply Noise and CTF',
                      help='If not selected, noise and CTF will not be simulated. If selected, noise and CTF will be applied'
                           ' then the the projections are CTF phase inverted.')
        form.addParam('targetSNR', params.FloatParam, default=0.1,
                      condition='noiseCTFChoice==%d' % NOISE_CTF_YES,
                      label='Tagret SNR',
                      help='The signal to noise ratio for the noise')
        form.addParam('ctfVoltage', params.FloatParam, default=200.0,
                      condition='noiseCTFChoice==%d' % NOISE_CTF_YES,
                      label='CTF Voltage',
                      help='Simualted voltage in the microscope in kV')
        form.addParam('ctfSphericalAberration', params.FloatParam, default=2,
                      condition='noiseCTFChoice==%d' % NOISE_CTF_YES,
                      label='CTF Spherical Aberration',
                      help='Microscope attribute')
        form.addParam('ctfMagnification', params.FloatParam, default=50000.0,
                      condition='noiseCTFChoice==%d' % NOISE_CTF_YES,
                      label='CTF Magnification',
                      help='Maginification in the microscope')
        form.addParam('ctfDefocusU', params.FloatParam, default=-5000.0,
                      condition='noiseCTFChoice==%d' % NOISE_CTF_YES,
                      label='CTF DefocusU',
                      help='defocus value (keep the value for DefocusV for simplicity)')
        form.addParam('ctfDefocusV', params.FloatParam, default=-5000.0,
                      condition='noiseCTFChoice==%d' % NOISE_CTF_YES,
                      label='CTF DefocusV',
                      help='defocus value (keep the value for DefocusU for simplicity)')
        form.addParam('ctfQ0', params.FloatParam, default=-0.112762,
                      condition='noiseCTFChoice==%d' % NOISE_CTF_YES,
                      label='CTF Q0',
                      help='Microscope attribute')

        form.addSection(label='Low pass filtering')
        form.addParam('lowPassChoice', params.EnumParam, default=LOWPASS_NO,
                      choices=['Add extra dose accumulation', 'Stay with only CTF'],
                      label='Use low pass filtering',
                      help='The generated volumes will be low pass filtered before projection to images.'
                           ' This simulates extra distortions similar to dose accumulation.'
                           ' However, CTF will already have such an effect.')
        line = form.addLine('Frequency (normalized)',
                            condition='lowPassChoice==%d' % LOWPASS_YES,
                            help='The cufoff frequency and raised coside width of the low pass filter.'
                                 ' For details: see "xmipp_transform_filter --fourier low_pass"')
        line.addParam('w1', params.FloatParam, default=0.25,
                      condition='lowPassChoice==%d' % LOWPASS_YES,
                      label='Cutoff frequency (0 -> 0.5)')
        line.addParam('raisedw', params.FloatParam, default=0.02,
                      condition='lowPassChoice==%d' % LOWPASS_YES,
                      label='Raised cosine width')

        form.addSection('Rigid body variability')
        form.addParam('rotationShiftChoice', params.EnumParam, default=ROTATION_SHIFT_YES,
                      choices=['Yes', 'No'],
                      label='Simulate Rotations and Shifts',
                      help='If yes, the images will be rotated and shifted randomly in the range,'
                           ' otherwise no rotations nor shifts will be applied')
        form.addParam('shiftx', params.EnumParam, label='Shift-X distribution',
                      condition='rotationShiftChoice==%d' % ROTATION_SHIFT_YES,
                      choices=['Uniform distribution', 'Gaussian distribution'],
                      default=ROTATION_UNIFORM,
                      display=params.EnumParam.DISPLAY_COMBO,
                      help='The distribution of random values')
        groupx_uni = form.addGroup('Shift-X Uniform distribution', condition='rotationShiftChoice==%d and shiftx==%d'
                                                                             % (ROTATION_SHIFT_YES,ROTATION_UNIFORM))
        groupx_uni.addParam('LowX', params.FloatParam, default=-5.0,
                      label='Lower value')
        groupx_uni.addParam('HighX', params.FloatParam, default=5.0,
                      label='Higher value')
        groupx_gas = form.addGroup('Shift-X Gaussian distribution', condition='rotationShiftChoice==%d and shiftx==%d'
                                                                             % (ROTATION_SHIFT_YES,ROTATION_GAUSS))
        groupx_gas.addParam('MeanX', params.FloatParam, default=0.0,
                      label='Mean value for the Gaussian distribution')
        groupx_gas.addParam('StdX', params.FloatParam, default=2.0,
                      label='Standard deviation for the Gaussian distribution')

        form.addParam('shifty', params.EnumParam, label='Shift-Y distribution',
                      condition='rotationShiftChoice==%d' % ROTATION_SHIFT_YES,
                      choices=['Uniform distribution', 'Gaussian distribution'],
                      default=ROTATION_UNIFORM,
                      display=params.EnumParam.DISPLAY_COMBO,
                      help='The distribution of random values')
        group = form.addGroup('Shift-Y Uniform distribution', condition='rotationShiftChoice==%d and shifty==%d'
                                                                             % (ROTATION_SHIFT_YES,ROTATION_UNIFORM))
        group.addParam('LowY', params.FloatParam, default=-5.0,
                      label='Lower value')
        group.addParam('HighY', params.FloatParam, default=5.0,
                      label='Higher value')
        group = form.addGroup('Shift-Y Gaussian distribution', condition='rotationShiftChoice==%d and shifty==%d'
                                                                             % (ROTATION_SHIFT_YES,ROTATION_GAUSS))
        group.addParam('MeanY', params.FloatParam, default=0.0,
                      label='Mean value for the Gaussian distribution')
        group.addParam('StdY', params.FloatParam, default=2.0,
                      label='Standard deviation for the Gaussian distribution')

        form.addParam('rot', params.EnumParam, label='Rot angle distribution',
                      condition='rotationShiftChoice==%d' % ROTATION_SHIFT_YES,
                      choices=['Uniform distribution', 'Gaussian distribution'],
                      default=ROTATION_UNIFORM,
                      display=params.EnumParam.DISPLAY_COMBO,
                      help='The distribution of random values')
        group = form.addGroup('Rot angle Uniform distribution', condition='rotationShiftChoice==%d and rot==%d'
                                                                             % (ROTATION_SHIFT_YES,ROTATION_UNIFORM))
        group.addParam('LowRot', params.FloatParam, default=0.0,
                      label='Lower value')
        group.addParam('HighRot', params.FloatParam, default=360.0,
                      label='Higher value')
        group = form.addGroup('Rot angle Gaussian distribution', condition='rotationShiftChoice==%d and rot==%d'
                                                                             % (ROTATION_SHIFT_YES,ROTATION_GAUSS))
        group.addParam('MeanRot', params.FloatParam, default=180.0,
                      label='Mean value for the Gaussian distribution')
        group.addParam('StdRot', params.FloatParam, default=90.0,
                      label='Standard deviation for the Gaussian distribution')

        form.addParam('tilt', params.EnumParam, label='Tilt angle distribution',
                      condition='rotationShiftChoice==%d' % ROTATION_SHIFT_YES,
                      choices=['Uniform distribution', 'Gaussian distribution'],
                      default=ROTATION_UNIFORM,
                      display=params.EnumParam.DISPLAY_COMBO,
                      help='The distribution of random values')
        group = form.addGroup('Tilt angle Uniform distribution', condition='rotationShiftChoice==%d and tilt==%d'
                                                                             % (ROTATION_SHIFT_YES,ROTATION_UNIFORM))
        group.addParam('LowTilt', params.FloatParam, default=0.0,
                      label='Lower value')
        group.addParam('HighTilt', params.FloatParam, default=180.0,
                      label='Higher value')
        group = form.addGroup('Tilt angle Gaussian distribution', condition='rotationShiftChoice==%d and tilt==%d'
                                                                             % (ROTATION_SHIFT_YES,ROTATION_GAUSS))
        group.addParam('MeanTilt', params.FloatParam, default=90.0,
                      label='Mean value for the Gaussian distribution')
        group.addParam('StdTilt', params.FloatParam, default=45.0,
                      label='Standard deviation for the Gaussian distribution')

        form.addParam('psi', params.EnumParam, label='Psi angle distribution',
                      condition='rotationShiftChoice==%d' % ROTATION_SHIFT_YES,
                      choices=['Uniform distribution', 'Gaussian distribution'],
                      default=ROTATION_UNIFORM,
                      display=params.EnumParam.DISPLAY_COMBO,
                      help='The distribution of random values')
        group = form.addGroup('Psi angle Uniform distribution', condition='rotationShiftChoice==%d and psi==%d'
                                                                             % (ROTATION_SHIFT_YES,ROTATION_UNIFORM))
        group.addParam('LowPsi', params.FloatParam, default=0.0,
                      label='Lower value')
        group.addParam('HighPsi', params.FloatParam, default=360.0,
                      label='Higher value')
        group = form.addGroup('Psi angle Gaussian distribution', condition='rotationShiftChoice==%d and psi==%d'
                                                                             % (ROTATION_SHIFT_YES,ROTATION_GAUSS))
        group.addParam('MeanPsi', params.FloatParam, default=180.0,
                      label='Mean value for the Gaussian distribution')
        group.addParam('StdPsi', params.FloatParam, default=90.0,
                      label='Standard deviation for the Gaussian distribution')
        # form.addParallelSection(threads=0, mpi=8)

        # --------------------------- INSERT steps functions --------------------------------------------

    def getInputPdb(self):
        """ Return the Pdb object associated with the normal modes. """
        return self.inputModes.get().getPdb()

    def _insertAllSteps(self):
        if(self.seedOption.get()):
            np.random.seed(int(time.time()))
        else:
            np.random.seed(0)
        if(self.confVar.get()==NMA_YES):
            if(self.importPdbs.get()):
                self._insertFunctionStep("copy_deformations")
            else:
                self._insertFunctionStep("generate_deformations")
            self._insertFunctionStep("generate_volume_from_pdb")
            if self.rotationShiftChoice == ROTATION_SHIFT_YES:
                self._insertFunctionStep("generate_rotation_and_shift")
            if self.rotationShiftChoice == ROTATION_SHIFT_NO:
                self._insertFunctionStep("generate_zero_rotation_and_shift")
            self._insertFunctionStep("project_volumes")
            if self.noiseCTFChoice == NOISE_CTF_YES:
                self._insertFunctionStep("apply_noise_and_ctf")
        else:
            self._insertFunctionStep("generate_links_to_volume")
            if self.rotationShiftChoice == ROTATION_SHIFT_YES:
                self._insertFunctionStep("generate_rotation_and_shift")
            if self.rotationShiftChoice == ROTATION_SHIFT_NO:
                self._insertFunctionStep("generate_zero_rotation_and_shift")
            self._insertFunctionStep("project_volumes")
            if self.noiseCTFChoice == NOISE_CTF_YES:
                self._insertFunctionStep("apply_noise_and_ctf")
        self._insertFunctionStep('createOutputStep')

    # --------------------------- STEPS functions --------------------------------------------
    def generate_deformations(self):
        # Find the right PDB file to use for data synthesis
        pdb_name1 = os.path.dirname(self.inputModes.get().getFileName()) + '/atoms.pdb'
        pdb_name2 = os.path.dirname(self.inputModes.get().getFileName()) + '/pseudoatoms.pdb'
        if os.path.exists(pdb_name1):
            fnPDB = pdb_name1
        else:
            fnPDB = pdb_name2
        # fnPDB = self.inputModes.get().getPdb().getFileName()
        # use the input relationship between the modes to generate normal mode amplitudes metadata
        fnModeList = replaceExt(self.inputModes.get().getFileName(),'xmd')
        modeAmplitude = self.modesAmplitudeRange.get()
        meshRowPoints = self.meshRowPoints.get()
        numberOfModes = self.inputModes.get().getSize()
        modeSelection = np.array(getListFromRangeString(self.modeList.get()))
        deformationFile = self._getExtraPath('GroundTruth.xmd')
        subtomogramMD = md.MetaData()
        # these vairables for alternating the generation when 3 clusters are selected
        cluster1 = cluster2 = cluster3 = False
        # these variables for the generaton of a mesh if selected
        XX, YY = np.meshgrid(np.linspace(start=-modeAmplitude, stop=modeAmplitude, num=meshRowPoints),
                             np.linspace(start=modeAmplitude, stop=-modeAmplitude, num=meshRowPoints))
        mode7_samples = XX.reshape(-1)
        mode8_samples = YY.reshape(-1)

        # iterate over the number of outputs (if mesh, this has to be calculated)
        numberOfVolumes = self.get_number_of_volumes()

        for i in range(numberOfVolumes):
            deformations = np.zeros(numberOfModes)

            if self.modeRelationChoice == MODE_RELATION_LINEAR:
                amplitude = self.modesAmplitudeRange.get()
                deformations[modeSelection - 1] = np.ones(len(modeSelection))*np.random.uniform(-amplitude, amplitude)
            elif self.modeRelationChoice == MODE_RELATION_3CLUSTERS:
                center_point = self.centerPoint.get()
                center1 = (-center_point, 0)
                center2 = (center_point, 0)
                center3 = (0, center_point)
                if not(cluster1 or cluster2 or cluster3):
                    cluster1 = True
                if cluster3:
                    deformations[modeSelection - 1] = center3
                    cluster3 = False
                if cluster2:
                    deformations[modeSelection - 1] = center2
                    cluster2 = False
                    cluster3 = True
                if cluster1:
                    deformations[modeSelection - 1] = center1
                    cluster1 = False
                    cluster2 = True
            elif self.modeRelationChoice == MODE_RELATION_RANDOM:
                amplitude=self.modesAmplitudeRange.get()
                deformations[modeSelection-1] = np.random.uniform(-amplitude, amplitude, len(modeSelection))
            elif self.modeRelationChoice == MODE_RELATION_PARABOLA:
                amplitude=self.modesAmplitudeRange.get()
                rv = np.random.uniform(0,1)
                point = (amplitude*cos(rv*pi), amplitude*sin(rv*pi))
                deformations[modeSelection-1] = point

            elif self.modeRelationChoice == MODE_RELATION_MESH:
                new_point=(mode7_samples[i],mode8_samples[i])
                deformations[modeSelection - 1] = new_point

            # we won't keep the first 6 modes
            deformations = deformations[6:]

            # params = " --pdb " + fnPDB
            # params+= " --nma " + fnModeList
            # params+= " -o " + self._getExtraPath(str(i+1).zfill(5)+'_df.pdb')
            # params+= " --deformations " + ' '.join(str(i) for i in deformations)
            # runProgram('xmipp_pdb_nma_deform', params)
            self.nma_deform_pdb(fnPDB,fnModeList,self._getExtraPath(str(i+1).zfill(5)+'_df.pdb'),deformations)

            subtomogramMD.setValue(md.MDL_IMAGE, self._getExtraPath(str(i+1).zfill(5)+'_projected'+'.spi'), subtomogramMD.addObject())
            subtomogramMD.setValue(md.MDL_NMA, list(deformations), i+1)

        subtomogramMD.write(deformationFile)

    def copy_deformations(self):
        pdbs_list = [f for f in glob.glob(self.pdbs_path.get())]
        # print(pdbs_list)
        # saving the list
        dump(pdbs_list, self._getExtraPath('pdb_list.pkl'))
        subtomogramMD = md.MetaData()
        i = 0
        for pdbfn in pdbs_list:
            i += 1
            createLink(pdbfn, self._getExtraPath(str(i).zfill(5)+'_df.pdb'))
            subtomogramMD.setValue(md.MDL_IMAGE, self._getExtraPath(str(i).zfill(5)+'_subtomogram'+'.vol'),
                                   subtomogramMD.addObject())
        subtomogramMD.write(self._getExtraPath('GroundTruth.xmd'))

    def generate_volume_from_pdb(self):
        numberOfVolumes = self.get_number_of_volumes()

        for i in range(numberOfVolumes):
            params = " -i " + self._getExtraPath(str(i + 1).zfill(5) + '_df.pdb')
            params += " --sampling " + str(self.samplingRate.get())
            params += " --size " + str(self.volumeSize.get())
            params += " -v 0 --centerPDB "
            runProgram('xmipp_volume_from_pdb', params)

        if self.lowPassChoice.get() is LOWPASS_YES:
            cutoff = self.w1.get()
            raisedw = self.raisedw.get()
            for i in range(numberOfVolumes):
                params = " -i " + self._getExtraPath(str(i + 1).zfill(5) + '_df.vol')
                params += " --fourier low_pass " + str(cutoff) + ' ' + str(raisedw)
                runProgram('xmipp_transform_filter', params)

    def generate_rotation_and_shift(self):
        subtomogramMD = md.MetaData(self._getExtraPath('GroundTruth.xmd'))
        numberOfVolumes = self.get_number_of_volumes()


        for i in range(numberOfVolumes):
            if (self.shiftx.get()==ROTATION_UNIFORM):
                shift_x1 = np.random.uniform(self.LowX.get(), self.HighX.get())
            else:
                shift_x1 = np.random.normal(self.MeanX.get(), self.StdX.get())
            if (self.shifty.get()==ROTATION_UNIFORM):
                shift_y1 = np.random.uniform(self.LowY.get(), self.HighY.get())
            else:
                shift_y1 = np.random.normal(self.MeanY.get(), self.StdY.get())
            if (self.rot.get()==ROTATION_UNIFORM):
                rot1 = np.random.uniform(self.LowRot.get(), self.HighRot.get())
            else:
                rot1 = np.random.normal(self.MeanRot.get(), self.StdRot.get())
            if (self.tilt.get()==ROTATION_UNIFORM):
                tilt1 = np.random.uniform(self.LowTilt.get(), self.HighTilt.get())
            else:
                tilt1 = np.random.normal(self.MeanTilt.get(), self.StdTilt.get())
            if (self.psi.get()==ROTATION_UNIFORM):
                psi1 = np.random.uniform(self.LowPsi.get(), self.HighPsi.get())
            else:
                psi1 = np.random.normal(self.MeanPsi.get(), self.StdPsi.get())

            # uniform over the sphere
            if (self.psi.get() == ROTATION_UNIFORM) and\
                (self.tilt.get()==ROTATION_UNIFORM) and \
                    (self.rot.get()==ROTATION_UNIFORM) and \
                    self.LowRot.get() == 0.0 and self.HighRot.get() == 360.0 and \
                    self.LowTilt.get() == 0.0 and self.HighTilt.get() == 180.0 and \
                    self.LowPsi.get() == 0.0 and self.HighPsi.get() == 360.0:
                x1,x2,x3 = np.random.uniform(0,1,3)
                R = np.array([
                    [np.cos(2*np.pi*x1), np.sin(2*np.pi*x1), 0],
                    [-np.sin(2*np.pi*x1), np.cos(2*np.pi*x1), 0],
                    [0, 0, 1]
                ])
                v = np.array([[np.cos(2*np.pi*x2)*np.sqrt(x3),
                              np.sin(2*np.pi*x2)*np.sqrt(x3),
                              np.sqrt(1-x3)]])
                H = np.eye(3) - 2*np.dot(v.T,v)
                M = -np.dot(H,R)
                trans_mat = np.zeros((4,4))
                trans_mat[:3,:3] = M
                rot1,tilt1,psi1,_,_,_ = matrix2eulerAngles(trans_mat)

            subtomogramMD.setValue(md.MDL_SHIFT_X, shift_x1, i + 1)
            subtomogramMD.setValue(md.MDL_SHIFT_Y, shift_y1, i + 1)
            subtomogramMD.setValue(md.MDL_ANGLE_ROT, rot1, i + 1)
            subtomogramMD.setValue(md.MDL_ANGLE_TILT, tilt1, i + 1)
            subtomogramMD.setValue(md.MDL_ANGLE_PSI, psi1, i + 1)
        subtomogramMD.write(self._getExtraPath('GroundTruth.xmd'))

    def generate_zero_rotation_and_shift(self):
        subtomogramMD = md.MetaData(self._getExtraPath('GroundTruth.xmd'))
        numberOfVolumes = self.get_number_of_volumes()

        for i in range(numberOfVolumes):
            rot1 = 0.0
            tilt1 = 0.0
            psi1 = 0.0
            shift_x1 = 0.0
            shift_y1 = 0.0
            subtomogramMD.setValue(md.MDL_SHIFT_X, shift_x1, i + 1)
            subtomogramMD.setValue(md.MDL_SHIFT_Y, shift_y1, i + 1)
            subtomogramMD.setValue(md.MDL_ANGLE_ROT, rot1, i + 1)
            subtomogramMD.setValue(md.MDL_ANGLE_TILT, tilt1, i + 1)
            subtomogramMD.setValue(md.MDL_ANGLE_PSI, psi1, i + 1)
        subtomogramMD.write(self._getExtraPath('GroundTruth.xmd'))


    def project_volumes(self):
        numberOfVolumes = self.get_number_of_volumes()

        sizeX = self.volumeSize.get()
        sizeY = self.volumeSize.get()
        volumeName = "_df.vol"
        subtomogramMD = md.MetaData(self._getExtraPath('GroundTruth.xmd'))
        for i in range(numberOfVolumes):
            rot = subtomogramMD.getValue(md.MDL_ANGLE_ROT, i+1)
            tilt = subtomogramMD.getValue(md.MDL_ANGLE_TILT, i+1)
            psi = subtomogramMD.getValue(md.MDL_ANGLE_PSI, i+1)
            x = subtomogramMD.getValue(md.MDL_SHIFT_X, i+1)
            y = subtomogramMD.getValue(md.MDL_SHIFT_Y, i+1)

            params = " -i " +  self._getExtraPath(str(i + 1).zfill(5) + volumeName)
            params += " -o " + self._getExtraPath(str(i + 1).zfill(5) + '_projected.spi')
            params += " --angles " + str(rot) + ' ' + str(tilt) + ' ' + str(psi) + ' ' + str(x) + ' ' + str(y)
            runProgram('xmipp_phantom_project', params)

    def apply_noise_and_ctf(self):
        numberOfVolumes = self.get_number_of_volumes()

        with open(self._getExtraPath('ctf.param'), 'a') as file:
            file.write(
                "\n".join([
                    "# XMIPP_STAR_1 *",
                    "data_noname",
                    "_ctfVoltage " + str(self.ctfVoltage.get()),
                    "_ctfSphericalAberration " + str(self.ctfSphericalAberration.get()),
                    "_ctfSamplingRate " + str(self.samplingRate.get()),
                    "_magnification " + str(self.ctfMagnification.get()),
                    "_ctfDefocusU " + str(self.ctfDefocusU.get()),
                    "_ctfDefocusV " + str(self.ctfDefocusV.get()),
                    "_ctfQ0 " + str(self.ctfQ0.get())]))
        #
        for i in range(numberOfVolumes):
            params = " -i " + self._getExtraPath(str(i + 1).zfill(5) + '_projected.spi')
            params += " --ctf " + self._getExtraPath('ctf.param')
            paramsNoiseCTF = params+ " --after_ctf_noise --targetSNR " + str(self.targetSNR.get())
            runProgram('xmipp_phantom_simulate_microscope', paramsNoiseCTF)

            # Phase flip:
            img_name = self._getExtraPath(str(i + 1).zfill(5)) + '_projected.spi'
            params_j = " -i " + img_name + " -o " + img_name
            params_j += " --ctf " + self._getExtraPath('ctf.param')
            runProgram('xmipp_ctf_phase_flip', params_j)

    def nma_deform_pdb(self, fnPDB, fnModeList, fnOut, deformList):
        def readPDB(fnIn):
            with open(fnIn) as f:
                lines = f.readlines()
            newlines = []
            for line in lines:
                if line.startswith("ATOM "):
                    try:
                        x = float(line[30:38])
                        y = float(line[38:46])
                        z = float(line[46:54])
                        newline = [x, y, z]
                        newlines.append(newline)
                    except:
                        pass
            return newlines

        def readModes(fnIn):
            modesMD = md.MetaData(fnIn)
            vectors = []
            for objId in modesMD:
                vecFn = modesMD.getValue(md.MDL_NMA_MODEFILE, objId)
                vec = np.loadtxt(vecFn)
                vectors.append(vec)
            return vectors

        def savePDB(list, fn, fn_original):
            with open(fn_original) as f:
                lines = f.readlines()
            newLines = []
            i = 0
            for line in lines:
                if line.startswith("ATOM "):
                    try:
                        x = list[i][0]
                        y = list[i][1]
                        z = list[i][2]
                        newLine = line[0:30] + "%8.3f%8.3f%8.3f" % (x, y, z) + line[54:]
                        i += 1
                    except:
                        pass
                else:
                    newLine = line
                newLines.append(newLine)
            with open(fn, mode='w') as f:
                f.writelines(newLines)
            pass

        pdb_array = np.array(readPDB(fnPDB))
        modes = readModes(fnModeList)
        for i in range(len(deformList)):
            pdb_array += deformList[i] * modes[7-1+i]
        savePDB(pdb_array, fnOut, fnPDB)

    def generate_links_to_volume(self):
        fn_volume = self._getExtraPath('reference')
        if(self.refAtomic.get()):
            pdbFn = self.refAtomic.get().getFileName()
            if getExt(pdbFn) == ".cif":
                pdbFn2 = replaceBaseExt(pdbFn, 'pdb')
                cifToPdb(pdbFn, pdbFn2)
                pdbFn = pdbFn2
            params = " -i " + pdbFn
            params += " -o " + fn_volume
            params += " --sampling " + str(self.samplingRate.get())
            params += " --size " + str(self.volumeSize.get())
            params += " -v 0 --centerPDB "
            runProgram('xmipp_volume_from_pdb', params)
        else:
            # Convert to spider format in case it is MRC
            params = "-i " + self.refVolume.get().getFileName()
            params += " -o " + fn_volume + ".vol --type vol"
            runProgram('xmipp_image_convert', params)
            # print(self.refVolume.get().getFileName())

        numberOfVolumes = self.get_number_of_volumes()

        deformationFile = self._getExtraPath('GroundTruth.xmd')
        imagesMD = md.MetaData()
        for i in range(numberOfVolumes):
            Vol_i = self._getExtraPath(str(i + 1).zfill(5) + '_df.vol')
            createLink(fn_volume+'.vol',Vol_i)
            imagesMD.setValue(md.MDL_IMAGE, self._getExtraPath(str(i + 1).zfill(5) + '_projected' + '.spi'),
                                   imagesMD.addObject())

        imagesMD.write(deformationFile)

    def createOutputStep(self):
        # first making a metadata for only the images:
        out_mdfn = self._getExtraPath('images.xmd')
        pattern = '"' + self._getExtraPath() + '/*_projected.spi"'
        command = '-p ' + pattern + ' -o ' + out_mdfn
        runProgram('xmipp_metadata_selfile_create', command)
        # now creating the output set of images as output:
        partSet = self._createSetOfParticles('images')
        xmipp3.convert.readSetOfParticles(self._getExtraPath('GroundTruth.xmd'), partSet)
        if (self.refVolume.get()):
            sr = self.refVolume.get().getSamplingRate()
        else:
            sr = self.samplingRate.get()
        partSet.setSamplingRate(sr)
        self._defineOutputs(outputImages=partSet)



    def get_number_of_volumes(self):
        if(self.importPdbs.get()):
            numberOfVolumes = len(glob.glob(self.pdbs_path.get()))
        elif self.modeRelationChoice.get() is MODE_RELATION_MESH:
            numberOfVolumes = self.meshRowPoints.get()*self.meshRowPoints.get()
        else:
            numberOfVolumes = self.numberOfVolumes.get()
        return numberOfVolumes

    # --------------------------- UTILS functions --------------------------------------------
    def _printWarnings(self, *lines):
        """ Print some warning lines to 'warnings.xmd', 
        the function should be called inside the working dir."""
        fWarn = open("warnings.xmd", 'w')
        for l in lines:
            print >> fWarn, l
        fWarn.close()

    def _getLocalModesFn(self):
        modesFn = self.inputModes.get().getFileName()
        return self._getBasePath(modesFn)

    # --------------------------- INFO functions --------------------------------------------
    def _summary(self):
        summary = []
        return summary

    def _validate(self):
        errors = []
        return errors

    def _citations(self):
        return ['harastani2022continuousflex']

    def _methods(self):
        pass


import numpy as np
from math import cos, sin, atan2, acos, radians, degrees
import torch
def eul2quat(arr, i):
    rot = radians(arr[i,0])
    tilt = radians(arr[i,1])
    psi = radians(arr[i,2])

    quaternion = [(cos(rot/2)*cos(tilt/2)*cos(psi/2))-(sin(rot/2)*cos(tilt/2)*sin(psi/2)),
                  (cos(rot/2)*sin(tilt/2)*sin(psi/2))-(sin(rot/2)*sin(tilt/2)*cos(psi/2)),
                  (cos(rot/2)*sin(tilt/2)*cos(psi/2))+(sin(rot/2)*sin(tilt/2)*sin(psi/2)),
                  (cos(rot/2)*cos(tilt/2)*sin(psi/2))+(sin(rot/2)*cos(tilt/2)*cos(psi/2))]
    return quaternion



def quater2euler(arr):
    qw = arr[0] 
    qx = arr[1]
    qy = arr[2]
    qz = arr[3]
    
    tilt = (qw**2)-(qx**2)-(qy**2)+(qz**2)
    if tilt > 1:
        tilt = 1.0
    elif tilt <-1:
        tilt = -1.0
    else:
        pass 
    
    euler = [degrees(atan2(2*((qy*qz)-(qw*qx)),2*((qx*qz)+(qw*qy)))),
             degrees(acos(tilt)),
             degrees(atan2(2*((qy*qz)+(qw*qx)),-2*((qx*qz)-(qw*qy))))]
    
    return euler


def quat2rotm(arr):

    q0 = arr[0] 
    q1 = arr[1]
    q2 = arr[2]
    q3 = arr[3]

    I = np.array([[1,0,0],[0,1,0],[0,0,1]])
    A = np.array([[0, -q3, q2],[q3, 0, -q1],[-q2, q1, 0]])

    rot_mat = I + 2 * q0 * A + 2 * A * A
    return rot_mat

"""
a = np.array([[134.81444 , 131.236146, 356.849227],[90, 90, 90]], dtype='float32')    

example = eul2quat(a,0)
print("quaternions ",example)


example2 = quater2euler(example)

print("angles in degrees",example2)
"""




# By Mohamad Harastani

import matplotlib.pyplot as plt
import numpy as np
from scipy.signal import resample_poly


def plot_quiver_2d(flow, title='optical flow'):
    u = flow[:, :, 0]
    v = flow[:, :, 1]
    num_rows = np.shape(flow)[0]
    num_cols = np.shape(flow)[1]
    x = np.arange(0, num_rows, 1)
    y = np.arange(0, num_cols, 1)
    y_pos, x_pos = np.meshgrid(y, x)
    fig, ax = plt.subplots()
    ax.quiver(y_pos, x_pos, v, u)
    plt.gca().invert_yaxis()
    plt.title(title)
    plt.show()


def plot_quiver_3d(flow, downsample=2, title='3D optical flow', normalize=False):
    u = flow[0, :, :, :]
    v = flow[1, :, :, :]
    w = flow[2, :, :, :]
    num_x = np.shape(flow)[1]
    num_y = np.shape(flow)[2]
    num_z = np.shape(flow)[3]

    factors = [(1, downsample), (1, downsample), (1, downsample)]

    for k in range(3):
        u = resample_poly(u, factors[k][0], factors[k][1], axis=k)
        v = resample_poly(v, factors[k][0], factors[k][1], axis=k)
        w = resample_poly(w, factors[k][0], factors[k][1], axis=k)

    x = np.arange(0, int(num_x / downsample), 1)
    y = np.arange(0, int(num_y / downsample), 1)
    z = np.arange(0, int(num_z / downsample), 1)
    x_pos, y_pos, z_pos = np.meshgrid(x, y, z)
    fig = plt.figure()
    ax = fig.gca(projection='3d')
    ax.quiver3D(x_pos, y_pos, z_pos, u, v, w, length=0.1, normalize=normalize)
    plt.title(title)
    plt.show()



# By Mohamad Harastani and Remi Vuillemot

from umap import UMAP
import sys
from joblib import load, dump

def umap_run(n_component, n_neigbors, n_epocks, pkl_pdbs, pkl_pca, pkl_out):
    pdbs_matrix = load(pkl_pdbs)
    umap = UMAP(n_components=n_component, n_neighbors=n_neigbors, n_epochs=n_epocks).fit(pdbs_matrix)
    Y = umap.transform(pdbs_matrix)
    dump(umap, pkl_pca)
    dump(Y, pkl_out)

if __name__ == '__main__':
    umap_run(int(sys.argv[1]),
             int(sys.argv[2]),
             int(sys.argv[3]),
             sys.argv[4],
             sys.argv[5],
             sys.argv[6])
    sys.exit()



import numpy as np
from struct import pack
import torch
import multiprocessing as multiprocessing

def readPDB(fnIn):
    with open(fnIn) as f:
        lines = f.readlines()
    return lines


def PDB2List(lines):
    newlines = []
    for line in lines:
        if line.startswith("ATOM "):
            try:
                x = float(line[30:38])
                y = float(line[38:46])
                z = float(line[46:54])
                newline = [x, y, z]
                newlines.append(newline)
            except:
                pass
    return newlines


def pdb_to_array(fnIn):
    return torch.Tensor(PDB2List(readPDB(fnIn)))


def euler_matrix(rot, tilt, psi):
    from math import sin, cos, radians
    t1 = -torch.deg2rad(psi)
    t2 = -torch.deg2rad(tilt)
    t3 = -torch.deg2rad(rot)
    a11 = torch.cos(t1) * torch.cos(t2) * torch.cos(t3) - torch.sin(t1) * torch.sin(t3)
    a12 = -torch.cos(t3) * torch.sin(t1) - torch.cos(t1) * torch.cos(t2) * torch.sin(t3)
    a13 = torch.cos(t1) * torch.sin(t2)
    a21 = torch.cos(t1) * torch.sin(t3) + torch.cos(t2) * torch.cos(t3) * torch.sin(t1)
    a22 = torch.cos(t1) * torch.cos(t3) - torch.cos(t2) * torch.sin(t1) * torch.sin(t3)
    a23 = torch.sin(t1) * torch.sin(t2)
    a31 = -torch.cos(t3) * torch.sin(t2)
    a32 = torch.sin(t2) * torch.sin(t3)
    a33 = torch.cos(t2)
    T = torch.tensor([[a11, a12, a13], [a21, a22, a23], [a31, a32, a33]])
    return T


def projectPDBPixel(pdb_coordinates, s, t, sigma):
    # Auxilary function, not meant to be used standalone
    row_size, column_size = pdb_coordinates.shape
    ng = row_size
    # print(row_size,column_size)
    sum_of_gaussians = 0
    sl_v = pdb_coordinates[:,0]
    tl_v = pdb_coordinates[:,1]
    s_v = torch.ones_like(sl_v)*s
    t_v = torch.ones_like(tl_v)*t
    
    exp_arg = -((s_v-sl_v)**2 + (t_v-tl_v)**2)/(2*sigma**2)
    exp_vec = torch.exp(exp_arg)
    return torch.sum(exp_vec, dim = 2)
def projectPDB_NP(PDB, size, sampling_rate =1, sigma=1, rot=0, tilt=0, psi=0, shift_x=0, shift_y=0, shift_z=0):
    #T = torch.linalg.inv(euler_matrix(torch.tensor(rot, dtype=torch.float), torch.tensor(tilt, dtype=torch.float), torch.tensor(psi, dtype=torch.float)))
    #PDB = torch.matmul(PDB, T)/sampling_rate
    PDB = PDB/sampling_rate
    shifts = torch.ones_like(PDB)
    shifts[:, 0] = shifts[:, 0] * shift_x
    shifts[:, 1] = shifts[:, 1] * shift_y
    shifts[:, 2] = shifts[:, 2] * shift_z
    PDB += shifts
    
    PDB = (PDB - torch.mean(PDB) )
    
    projection = torch.zeros([size, size])
    
    limit = int(size/2)
    l = torch.arange(-limit, limit, 1)
    y, x = torch.meshgrid(l, l)
    x = x.unsqueeze(2)
    xx = x.repeat(1,1,PDB.shape[0])
    y = y.unsqueeze(2)
    yy = y.repeat(1,1,PDB.shape[0])
    res = projectPDBPixel(PDB, xx, yy, sigma)
    return res

def projectPDB2Image(PDB, size, sampling_rate =1, sigma=1, rot=0, tilt=0, psi=0, shift_x=0, shift_y=0, shift_z=0):
    # This function gives the same views as xmipp_phantom_project, however, it performs the projection by representing
    # atoms by 3D Gaussians

    # PDB: array of atomic coordinates
    # size: image size that you want
    # sampling rate: the sampling rate on the image
    # sigma: the Gaussian size for each atom (here the projection is with fixed Gaussian)
    # rot, tilt, psi: Euler angles of the view
    # shift_x, shift_y, shift_z: the shifting applied to the atomic structure before projection
    T = torch.linalg.inv(euler_matrix(torch.tensor(rot, dtype=torch.float), torch.tensor(tilt, dtype=torch.float), torch.tensor(psi, dtype=torch.float)))
    PDB = torch.matmul(PDB, T.to('cpu')) / sampling_rate
    shifts = torch.ones_like(PDB)
    shifts[:, 0] = shifts[:, 0] * shift_x
    shifts[:, 1] = shifts[:, 1] * shift_y
    shifts[:, 2] = shifts[:, 2] * shift_z
    PDB += shifts

    # projection = np.zeros([size, size])
    limit = int(size/2)
    l = torch.arange(-limit, limit, 1)
    x, y = torch.meshgrid(l, l)
    xv = torch.reshape(x, (-1,))
    yv = torch.reshape(y, (-1,))
    ps = [(xv[i], yv[i]) for i in range(len(xv))]
    
    global segment
    def segment(p):
        return projectPDBPixel(PDB, p[0], p[1], sigma)
    with multiprocessing.Pool(20) as pool:
        values = pool.map(segment,ps)
    values = torch.reshape(list(values),[size,size])
    return values


def PDBVoxel(pdb_coordinates,r, s, t, sigma, sampling_rate):
    # Auxilary function, not meant to be used standalone
    row_size, column_size = pdb_coordinates.shape
    ng = row_size
    # print(row_size,column_size)
    rl_v = pdb_coordinates[:,0]/sampling_rate
    sl_v = pdb_coordinates[:,1]/sampling_rate
    tl_v = pdb_coordinates[:,2]/sampling_rate
    r_v = np.ones_like(rl_v)*r
    s_v = np.ones_like(sl_v)*s
    t_v = np.ones_like(tl_v)*t
    exp_arg = -((r_v-rl_v)**2 + (s_v-sl_v)**2 + (t_v-tl_v)**2)/(2*sigma**2)
    exp_vec = np.exp(exp_arg)
    return np.sum(exp_vec)


def PDB2Volume(PDB, volume_size, sigma, sampling_rate):
    # This function gives the same as xmipp_volume_from_pdb, however, it represents atoms by 3D Gaussians

    # PDB: array of atomic coordinates
    # volume_size: output volume size
    # sampling rate: the sampling rate
    # sigma: the Gaussian size for each atom

    limit = int(volume_size/2)
    z, y, x = np.mgrid[-limit:limit, -limit:limit, -limit:limit]
    xv = np.reshape(x, -1)
    yv = np.reshape(y, -1)
    zv = np.reshape(z, -1)
    ps = [(xv[i], yv[i], zv[i]) for i in range(len(xv))]

    global segment

    def segment(p):
        return PDBVoxel(PDB, p[0], p[1], p[2], sigma, sampling_rate)
    with multiprocessing.Pool(processes = 8) as pool:
        values = pool.map(segment,ps)
    values = np.reshape(list(values),[volume_size,volume_size,volume_size])
    return values


def save_volume(vol, filename):
    vol = np.float32(vol)
    # From the spider format:
    labels = {i - 1: v for i, v in [
        (1, 'NZ'),
        (2, 'NY'),
        (3, 'IREC'),
        (5, 'IFORM'),
        (6, 'IMAMI'),
        (7, 'FMAX'),
        (8, 'FMIN'),
        (9, 'AV'),
        (10, 'SIG'),
        (12, 'NX'),
        (13, 'LABREC'),
        (14, 'IANGLE'),
        (15, 'PHI'),
        (16, 'THETA'),
        (17, 'GAMMA'),
        (18, 'XOFF'),
        (19, 'YOFF'),
        (20, 'ZOFF'),
        (21, 'SCALE'),
        (22, 'LABBYT'),
        (23, 'LENBYT'),
        (24, 'ISTACK/MAXINDX'),
        (26, 'MAXIM'),
        (27, 'IMGNUM'),
        (28, 'LASTINDX'),
        (31, 'KANGLE'),
        (32, 'PHI1'),
        (33, 'THETA1'),
        (34, 'PSI1'),
        (35, 'PHI2'),
        (36, 'THETA2'),
        (37, 'PSI2'),
        (38, 'PIXSIZ'),
        (39, 'EV'),
        (40, 'PROJ'),
        (41, 'MIC'),
        (42, 'NUM'),
        (43, 'GLONUM'),
        (101, 'PSI3'),
        (102, 'THETA3'),
        (103, 'PHI3'),
        (104, 'LANGLE')]}

    # Inverse.
    locations = {v: k for k, v in labels.items()}
    """Save volume vol into a file, with the spider format."""
    nx, ny, nz = vol.shape
    fields = [0.0] * nx
    values = {
        'NZ': nz, 'NY': ny,
        'IREC': 3,  # number of records (including header records)
        'IFORM': 3,  # 3D volume
        'FMAX': vol.max(), 'FMIN': vol.min(),
        'AV': vol.mean(), 'SIG': vol.std(),
        'NX': nx,
        'LABREC': 1,  # number of records in file header (label)
        'SCALE': 1,
        'LABBYT': 4 * nx,  # number of bytes in header
        'LENBYT': 4 * nx,  # record length in bytes (only 1 in our header)
    }
    for label, value in values.items():
        fields[locations[label]] = float(value)
    header = pack('%df' % nx, *fields)
    with open(filename, 'wb') as f:
        f.write(header)
        vol.tofile(f)
        



import glob

import numpy as np
from torch.utils.data import Dataset
from ..utils import spi2array, eul2quat, min_max
import torch
import pwem.emlib.metadata as md
class cryodata(Dataset):

    def __init__(self, path, output_path, flag='nma', mode = 'train', transform=None):
        self.path = path
        self.flag = flag
        self.mode = mode
        self.transform = transform
        mdImgs = md.MetaData(self.path)
        if mode == 'train':
            rot = []
            tilt = []
            psi = []
            nma = []
            shift_x = []
            shift_y = []
            imgPath = []
            for objId in mdImgs:
                imgPath.append(mdImgs.getValue(md.MDL_IMAGE, objId))
                rot.append(mdImgs.getValue(md.MDL_ANGLE_ROT, objId))
                tilt.append(mdImgs.getValue(md.MDL_ANGLE_TILT, objId))
                psi.append(mdImgs.getValue(md.MDL_ANGLE_PSI, objId))
                shift_x.append(mdImgs.getValue(md.MDL_SHIFT_X, objId))
                shift_y.append(mdImgs.getValue(md.MDL_SHIFT_Y, objId))
                nma.append(mdImgs.getValue(md.MDL_NMA, objId))
            self.images_Path = imgPath
            rot_ = torch.tensor(rot)
            tilt_ = torch.tensor(tilt)
            psi_ = torch.tensor(psi)
            shiftx = torch.tensor(shift_x)
            shifty = torch.tensor(shift_y)

            self.angles = torch.column_stack((rot_, tilt_, psi_))
            self.quaternions = torch.zeros((self.angles.shape[0], 4), dtype=torch.float32)
            for i in range(len(self.angles)):
                self.quaternions[i, :] = torch.tensor(eul2quat(self.angles, i))
            self.shifts, min_shf, max_shf = min_max(torch.column_stack((shiftx, shifty)))
            self.amplitudes, min_nma, max_nma = min_max(torch.tensor(nma, dtype=torch.float32))

            min_max_nma = torch.row_stack((min_nma, max_nma))
            min_max_shf = torch.row_stack((min_shf, max_shf))
            np.savetxt(output_path + '/min_max_nma.txt', min_max_nma.numpy())
            np.savetxt(output_path + '/min_max_shf.txt', min_max_shf.numpy())
        elif self.mode=='inference':
            imgPath = []
            for objId in mdImgs:
                imgPath.append(mdImgs.getValue(md.MDL_IMAGE, objId))
            self.images_Path = imgPath
        else:
            pass



    def __len__(self):
        return len(self.images_Path)

    def __getitem__(self, item):
        if self.mode == 'train':
            if self.flag == 'nma':
                amplitudes = self.amplitudes[item]
                image_name = self.images_Path[item]
                spi_array = spi2array(image_name)
                if self.transform:
                    spi_array = self.transform(spi_array)
                    amplitudes = torch.tensor(amplitudes)
                return spi_array, amplitudes
            elif self.flag == 'ang':
                angles = self.quaternions[item]
                image_name = self.images_Path[item]
                spi_array = spi2array(image_name)
                if self.transform:
                    spi_array = self.transform(spi_array)
                    angles = torch.tensor(angles) 
                return spi_array, angles, image_name
            elif self.flag == 'shf':
                shifts = self.shifts[item]    
                image_name = self.images_Path[item]
                spi_array = spi2array(image_name)
                if self.transform:
                    spi_array = self.transform(spi_array)
                    shifts = torch.tensor(shifts)
                return spi_array, shifts
            elif self.flag=='all':
                amplitudes = self.amplitudes[item]
                angles = self.quaternions[item]
                shifts = self.shifts[item]
                image_name = self.images_Path[item]
                spi_array = spi2array(image_name)
                if self.transform:
                    spi_array = self.transform(spi_array)
                    amplitudes = torch.tensor(amplitudes)
                    angles = torch.tensor(angles)
                    shifts = torch.tensor(shifts)
                    params = torch.cat([amplitudes, angles, shifts])
                return spi_array, params
        elif self.mode == 'inference':
            image_name = self.images_Path[item]
            spi_array = spi2array(image_name)
            if self.transform:
                spi_array = self.transform(spi_array)
            return spi_array, image_name



# **************************************************************************
# * Authors:  Mohamad Harastani          (mohamad.harastani@igbmc.fr)
# *           Remi Vuillemot             (remi.vuillemot@upmc.fr)
# * IMPMC, UPMC Sorbonne University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# **************************************************************************

import numpy as np
from pyworkflow.protocol.params import StringParam, LabelParam, EnumParam, FloatParam, PointerParam
from pyworkflow.viewer import (ProtocolViewer, DESKTOP_TKINTER, WEB_DJANGO)
from pwem.viewers import ChimeraView
from pwem.objects.data import SetOfParticles,SetOfVolumes
from continuousflex.viewers.nma_plotter import FlexNmaPlotter
from continuousflex.protocols import FlexProtDimredPdb
import matplotlib.pyplot as plt
from pwem.emlib.image import ImageHandler
from joblib import load
from continuousflex.viewers.tk_dimred import PCAWindowDimred
from continuousflex.protocols.data import Point, Data, PathData
from pwem.viewers import VmdView
from pyworkflow.utils.path import cleanPath, makePath
from continuousflex.protocols.utilities.genesis_utilities import numpyArr2dcd, dcd2numpyArr
from continuousflex.protocols.utilities.pdb_handler import ContinuousFlexPDBHandler
from pyworkflow.gui.browser import FileBrowserWindow
from continuousflex.protocols.protocol_pdb_dimred import REDUCE_METHOD_PCA, REDUCE_METHOD_UMAP
from continuousflex.protocols.protocol_batch_pdb_cluster import FlexBatchProtClusterSet
import os

X_LIMITS_NONE = 0
X_LIMITS = 1
Y_LIMITS_NONE = 0
Y_LIMITS = 1
Z_LIMITS_NONE = 0
Z_LIMITS = 1

ANIMATION_INV=0
ANIMATION_AVG=1
ANIMATION_PCA=2

NUM_POINTS_TRAJECTORY=10


class FlexProtPdbDimredViewer(ProtocolViewer):
    """ Visualization of dimensionality reduction on atomic structures
    """
    _label = 'viewer PDBs dimred'
    _targets = [FlexProtDimredPdb]
    _environments = [DESKTOP_TKINTER, WEB_DJANGO]

    def __init__(self, **kwargs):
        ProtocolViewer.__init__(self, **kwargs)
        self._data = None

    def _defineParams(self, form):
        form.addSection(label='Visualization')

        group = form.addGroup("Display Singular Values")
        group.addParam('displayPcaSingularValues', LabelParam,
                      label="Display singular values",
                      help="The values should help you see how many dimensions are in the data ",
                      condition=self.protocol.method.get()==REDUCE_METHOD_PCA)

        group = form.addGroup("Display PCA")
        group.addParam('displayPCA', LabelParam,
                      label='Display PCA axes',
                      help='Open a GUI to visualize the PCA space'
                           ' to draw and adjust trajectories.')

        group.addParam('pcaAxes', StringParam, default="1 2",
                       label='Axes to display' )

        group = form.addGroup("Animation tool")

        group.addParam('displayAnimationtool', LabelParam,
                      label='Open Animation tool ',
                      help='Open a GUI to analyze the PCA space'
                           ' to draw and adjust trajectories and create clusters.')

        group.addParam('inputSet', PointerParam, pointerClass ='SetOfParticles,SetOfVolumes',
                      label='(Optional) Em data for cluster animation',  allowsNull=True,
                      help="Provide a EM data set that match the PDB data set to visualize animation on 3D reconstructions")


        # form.addParam("dataSet", StringParam, default= "", label="Data set label")


        group = form.addGroup("Figure parameters")

        group.addParam('s', FloatParam, default=10, allowsNull=True,
                       label='Point radius')
        group.addParam('alpha', FloatParam, default=0.5, allowsNull=True,
                       label='Point transparancy')
        group.addParam('xlimits_mode', EnumParam,
                      choices=['Automatic (Recommended)', 'Set manually x-axis limits'],
                      default=X_LIMITS_NONE,
                      label='x-axis limits', display=EnumParam.DISPLAY_COMBO,
                      help='This allows you to use a specific range of x-axis limits')
        group.addParam('xlim_low', FloatParam, default=None,
                      condition='xlimits_mode==%d' % X_LIMITS,
                      label='Lower x-axis limit')
        group.addParam('xlim_high', FloatParam, default=None,
                      condition='xlimits_mode==%d' % X_LIMITS,
                      label='Upper x-axis limit')
        group.addParam('ylimits_mode', EnumParam,
                      choices=['Automatic (Recommended)', 'Set manually y-axis limits'],
                      default=Y_LIMITS_NONE,
                      label='y-axis limits', display=EnumParam.DISPLAY_COMBO,
                      help='This allows you to use a specific range of y-axis limits')
        group.addParam('ylim_low', FloatParam, default=None,
                      condition='ylimits_mode==%d' % Y_LIMITS,
                      label='Lower y-axis limit')
        group.addParam('ylim_high', FloatParam, default=None,
                      condition='ylimits_mode==%d' % Y_LIMITS,
                      label='Upper y-axis limit')
        group.addParam('zlimits_mode', EnumParam,
                      choices=['Automatic (Recommended)', 'Set manually z-axis limits'],
                      default=Z_LIMITS_NONE,
                      label='z-axis limits', display=EnumParam.DISPLAY_COMBO,
                      help='This allows you to use a specific range of z-axis limits')
        group.addParam('zlim_low', FloatParam, default=None,
                      condition='zlimits_mode==%d' % Z_LIMITS,
                      label='Lower z-axis limit')
        group.addParam('zlim_high', FloatParam, default=None,
                      condition='zlimits_mode==%d' % Z_LIMITS,
                      label='Upper z-axis limit')


    def _getVisualizeDict(self):
        return {
                'displayPCA': self._displayPCA,
                'displayAnimationtool': self._displayAnimationtool,
                'displayPcaSingularValues': self.viewPcaSinglularValues,
                }

    def _displayPCA(self, paramName):
        axes_str = str.split(self.pcaAxes.get())
        axes = []
        for i in axes_str : axes.append(int(i.strip()))

        dim = len(axes)
        if dim ==0 or dim >3:
            return self.errorMessage("Can not read input PCA axes selection", "Invalid Input")

        data = self.getData()
        plotter = FlexNmaPlotter(data= data,
                                      xlim_low=self.xlim_low.get(), xlim_high=self.xlim_high.get(),
                                      ylim_low=self.ylim_low.get(), ylim_high=self.ylim_high.get(),
                                      zlim_low=self.zlim_low.get(), zlim_high=self.zlim_high.get(),
                                      alpha=self.alpha, s=self.s, cbar_label=None)
        if dim == 1:
            data.XIND = axes[0]-1
            plotter.plotArray1D("PCA","%i component"%(axes[0]),"")
        if dim == 2:
            data.YIND = axes[1]-1
            plotter.plotArray2D_xy("PCA","%i component"%(axes[0]),"%i component"%(axes[1]))
        if dim == 3:
            data.ZIND = axes[2]-1
            plotter.plotArray3D_xyz("PCA","%i component"%(axes[0]),"%i component"%(axes[1]),"%i component"%(axes[2]))
        plotter.show()

    def _displayAnimationtool(self, paramName):
        self.trajectoriesWindow = self.tkWindow(PCAWindowDimred,
                                                title='Animation tool',
                                                dim=self.protocol.reducedDim.get(),
                                                data=self.getData(),
                                                callback=self._generateAnimation,
                                                loadCallback=self._loadAnimation,
                                                saveCallback=self._saveAnimation,
                                                saveClusterCallback=self.saveClusterCallback,
                                                numberOfPoints=5,
                                                limits_mode=0,
                                                LimitL=None,
                                                LimitH=None,
                                                xlim_low=self.xlim_low.get(),
                                                xlim_high=self.xlim_high.get(),
                                                ylim_low=self.ylim_low.get(),
                                                ylim_high=self.ylim_high.get(),
                                                zlim_low=self.zlim_low.get(),
                                                zlim_high=self.zlim_high.get(),
                                                s=self.s,
                                                alpha=self.alpha,
                                                cbar_label="Cluster")
        return [self.trajectoriesWindow]


    def viewPcaSinglularValues(self, paramName):
        pca = load(self.protocol._getExtraPath('pca_pickled.joblib'))
        fig = plt.figure('PCA singlular values')
        plt.stem(np.arange(1, len(pca.singular_values_)+1), pca.singular_values_)
        plt.show()
        pass

    def getData(self):
        if self._data is None:
            self._data = self.loadData()
        return self._data

    def loadData(self):
        data = Data()
        pdb_matrix = np.loadtxt(self.protocol.getOutputMatrixFile())

        # dataSet = self.dataSet.get().split(";")
        # n_data = len(dataSet)
        # if n_data >1:
        #     weights = []
        #     for i in range(n_data):
        #         if dataSet[i] != '':
        #             for j in range(int(dataSet[i])):
        #                 weights.append(i/n_data)
        #
        # else:
        #
        weights = [0 for i in range(pdb_matrix.shape[0])]

        for i in range(pdb_matrix.shape[0]):
            data.addPoint(Point(pointId=i+1, data=pdb_matrix[i, :],weight=weights[i]))
        return data

    def _generateAnimation(self):
        prot = self.protocol
        initPDB = ContinuousFlexPDBHandler(prot.getPDBRef())

        # Get animation root
        animation = self.trajectoriesWindow.getClusterName()
        animationPath = prot._getExtraPath('animation_%s' % animation)
        if not os.path.isdir(animationPath):
            cleanPath(animationPath)
            makePath(animationPath)
        animationRoot = os.path.join(animationPath, '')

        # get trajectory coordinates
        animtype = self.trajectoriesWindow.getAnimationType()
        coords_list = []
        if animtype ==ANIMATION_INV:
            trajectoryPoints = np.array([p.getData() for p in self.trajectoriesWindow.pathData])
            if trajectoryPoints.shape[0] == 0 :
                return self.trajectoriesWindow.showError("No animation to show.")
            np.savetxt(animationRoot + 'trajectory.txt', trajectoryPoints)
            pca = load(prot._getExtraPath('pca_pickled.joblib'))
            deformations = pca.inverse_transform(trajectoryPoints)
            for i in range(self.trajectoriesWindow.numberOfPoints):
                coords_list.append(deformations[i].reshape((initPDB.n_atoms, 3)))
        else :
            # read save coordinates
            coords = dcd2numpyArr(self.protocol._getExtraPath("coords.dcd"))

            # get class dict
            classDict = {}
            count = 0 #CLUSTERINGTAG
            for p in self.trajectoriesWindow.data:
                clsId = str(int(p._weight)) #CLUSTERINGTAG
                if clsId in classDict:
                    classDict[clsId].append(count)
                else:
                    classDict[clsId] = [count]
                count += 1

            if animtype == ANIMATION_AVG:
                # compute avg
                for i in classDict:
                    coord_avg = np.mean(coords[np.array(classDict[i])], axis=0)
                    coords_list.append(coord_avg.reshape((initPDB.n_atoms, 3)))


        # Generate DCD trajectory
        initdcdcp = initPDB.copy()
        initdcdcp.coords = coords_list[0]
        initdcdcp.write_pdb(animationRoot+"trajectory.pdb")
        numpyArr2dcd(arr = np.array(coords_list), filename=animationRoot+"trajectory.dcd")

        # Generate the vmd script
        vmdFn = animationRoot + 'trajectory.vmd'
        vmdFile = open(vmdFn, 'w')
        vmdFile.write("""
        mol new %strajectory.pdb waitfor all
        mol addfile %strajectory.dcd waitfor all
        animate style Rock
        display projection Orthographic
        mol modcolor 0 0 Index
        mol modstyle 0 0 Tube 1.000000 8.000000
        animate speed 0.75
        animate forward
        animate delete  beg 0 end 0 skip 0 0
        """ % (animationRoot,animationRoot))
        vmdFile.close()

        VmdView(' -e ' + vmdFn).show()

    def saveClusterCallback(self, tkWindow):
        # get cluster name
        clusterName = "animation_" + tkWindow.getClusterName()

        # get input metadata
        inputSet = self.inputSet.get()
        if inputSet is None:
            tkWindow.showError("Select an EM set before exporting clusters.")
            return

        classID=[]
        for p in tkWindow.data:
            classID.append(int(p._weight))

        if isinstance(inputSet, SetOfParticles):
            classSet = self.protocol._createSetOfClasses2D(inputSet, clusterName)
        else:
            classSet = self.protocol._createSetOfClasses3D(inputSet,clusterName)

        def updateItemCallback(item, row):
            item.setClassId(row)

        class itemDataIterator:
            def __init__(self, clsID):
                self.clsID = clsID

            def __iter__(self):
                self.n = 0
                return self

            def __next__(self):
                if self.n > len(self.clsID)-1:
                    return 0
                else:
                    index = self.clsID[self.n]
                    self.n += 1
                    return index

        classSet.classifyItems(
            updateItemCallback=updateItemCallback,
            updateClassCallback=None,
            itemDataIterator=iter(itemDataIterator(classID)),
            classifyDisabled=False,
            iterParams=None,
            doClone=True)

        # Run reconstruction
        self.protocol._defineOutputs(**{clusterName : classSet})
        project = self.protocol.getProject()
        newProt = project.newProtocol(FlexBatchProtClusterSet)
        newProt.setObjLabel(clusterName)
        newProt.inputSet.set(getattr(self.protocol, clusterName))
        project.launchProtocol(newProt)
        project.getRunsGraph()

        tkWindow.showInfo("Successfully exported clustering.")

    def _loadAnimation(self):
        browser = FileBrowserWindow("Select animation directory",
                                    self.getWindow(), self.protocol._getExtraPath(),
                                    onSelect=self._loadAnimationData)
        browser.show()

    def _loadAnimationData(self, obj):

        if not obj.isDir() :
            return self.trajectoriesWindow.showError('Not a directory')

        loaded = []
        trajPath = obj.getPath()
        trajFile = os.path.join(trajPath,'trajectory.txt')
        if os.path.isfile(trajFile) and os.path.getsize(trajFile) != 0:
            trajectoryPoints = np.loadtxt(trajFile)
            data = PathData(dim=trajectoryPoints.shape[1])
            for i, row in enumerate(trajectoryPoints):
                data.addPoint(Point(pointId=i + 1, data=list(row), weight=0))
            loaded.append("trajectory.txt")

        clusterFile = os.path.join(trajPath,'clusters.txt')
        if os.path.isfile(clusterFile) and os.path.getsize(clusterFile) != 0:
            clusterPoints = np.loadtxt(clusterFile)
            i=0
            for p in self.trajectoriesWindow.data:
                p._weight = clusterPoints[i]
                i+=1
            loaded.append("clusters.txt")
        if len(loaded) ==0:
            return self.trajectoriesWindow.showError('Animation files not found. ')
        else:
            self.trajectoriesWindow.showInfo('Successfully loaded : %s.' %str(loaded))

        self.trajectoriesWindow.setPathData(data)
        self.trajectoriesWindow._onUpdateClick()
        self.trajectoriesWindow._checkNumberOfPoints()

    def _saveAnimation(self, tkWindow):
        # get cluster name
        animationPath = self.protocol._getExtraPath("animation_" + tkWindow.getClusterName())
        if not os.path.isdir(animationPath):
            cleanPath(animationPath)
            makePath(animationPath)
        animationRoot = os.path.join(animationPath, '')
        trajectoryPoints = np.array([p.getData() for p in self.trajectoriesWindow.pathData])
        saved=[]
        if trajectoryPoints.shape[0] != 0:
            np.savetxt(animationRoot + 'trajectory.txt', trajectoryPoints)
            saved.append('trajectory.txt')

        classID=[]
        for p in tkWindow.data:
            classID.append(int(p._weight))
        if set(classID) != {0}:
            np.savetxt(animationRoot + 'clusters.txt', np.array(classID))
            saved.append('clusters.txt')

        try :
            self.trajectoriesWindow.plotter.figure.savefig(animationRoot + 'figure.png',dpi=500)
        except:
            pass

        if len(saved) != 0:
            self.trajectoriesWindow.showInfo('Successfully saved : %s.' % str(saved))
        else:
            self.trajectoriesWindow.showError('No animation state to save.')


class VolumeTrajectoryViewer(ProtocolViewer):
    """ Visualization of a SetOfVolumes as a trajectory with ChimeraX
    """
    _label = 'Volume trajectory viewer'
    _targets = [SetOfVolumes]

    def _defineParams(self, form):
        form.addSection(label='Visualization')
        form.addParam('displayTrajectories', LabelParam,
                      label='ChimeraX',
                      help='Open the trajectory in ChimeraX.')
    def _getVisualizeDict(self):
        return {
                'displayTrajectories': self._visualize,
                }

    def _visualize(self, obj, **kwargs):
        """visualisation for volumes set"""
        volNames = ""
        for i in self.protocol:
            i.setSamplingRate(self.protocol.getSamplingRate())
            vol = ImageHandler().read(i)
            volName = os.path.abspath(self._getPath("tmp%i.vol"%i.getObjId()))
            vol.write(volName)
            volNames += volName+" "
        # Show Chimera
        tmpChimeraFile = self._getPath("chimera.cxc")
        with open(tmpChimeraFile, "w") as f:
            f.write("open %s vseries true \n" % volNames)
            # f.write("volume #1 style surface level 0.5")
            f.write("vseries play #1 loop true maxFrameRate 7 direction oscillate \n")

        cv = ChimeraView(tmpChimeraFile)
        return [cv]



# By Mohamad Harastani

from spider_files3 import open_volume, save_volume

import time
import numpy as np
import sys
import os


def opflow_vols(path_vol0, path_vol1, pyr_scale, levels, winsize, iterations, poly_n, poly_sigma, factor1=100,
                factor2=100, path_volx='x_OF_3D.vol', path_voly='y_OF_3D.vol', path_volz='z_OF_3D.vol', gpu_id=0):
    # Convention here is in reverse order
    vol0 = open_volume(path_vol0)
    vol1 = open_volume(path_vol1)

    vol0 = vol0 * factor1
    vol1 = vol1 * factor2
    os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
    os.environ["CUDA_VISIBLE_DEVICES"] = str(gpu_id)

    import pycuda.autoinit
    import farneback3d

    optflow = farneback3d.Farneback(
        pyr_scale=pyr_scale,  # Scaling between multi-scale pyramid levels
        levels=levels,  # Number of multi-scale levels
        winsize=winsize,  # Window size for Gaussian filtering of polynomial coefficients
        num_iterations=iterations,  # Iterations on each multi-scale level
        poly_n=poly_n,  # Size of window for weighted least-square estimation of polynomial coefficients
        poly_sigma=poly_sigma,  # Sigma for Gaussian weighting of least-square estimation of polynomial coefficients
    )
    t0 = time.time()
    # perform OF:
    flow = optflow.calc_flow(vol0, vol1)
    t_end = time.time()
    print("spent on calculating 3D optical flow", np.floor((t_end - t0) / 60), "minutes and",
          np.round(t_end - t0 - np.floor((t_end - t0) / 60) * 60), "seconds")

    # Extracting the flows in x, y and z dimensions:
    Flowx = flow[0, :, :, :]
    Flowy = flow[1, :, :, :]
    Flowz = flow[2, :, :, :]

    save_volume(Flowx, path_volx)
    save_volume(Flowy, path_voly)
    save_volume(Flowz, path_volz)


if __name__ == '__main__':
    if len(sys.argv) < 9 or len(sys.argv) > 15:
        print('optical flow will not be calculated due to wrong arguments')
    else:
        opflow_vols(sys.argv[1],
                    sys.argv[2],
                    float(sys.argv[3]),
                    int(sys.argv[4]),
                    int(sys.argv[5]),
                    int(sys.argv[6]),
                    int(sys.argv[7]),
                    float(sys.argv[8]),
                    int(sys.argv[9]),
                    int(sys.argv[10]),
                    sys.argv[11],
                    sys.argv[12],
                    sys.argv[13],
                    int(sys.argv[14])
                    )
    sys.exit()


import torch.nn as nn
import torch
from . import ResNet, mlp, Bottleneck, BasicBlock
#from utils import projectPDB_NP, normalize, torch_normalize, quater2euler



class deephemnma(nn.Module):
    def __init__(self, output):
        super(deephemnma, self).__init__()
        self.output = output
        self.resnet = ResNet(BasicBlock, [3, 4, 6, 3])
        self.mlp = mlp(output)
    """
    def forward(self, x, pdb, mode = 'train'):
        resnet = self.resnet(x)
        flat = torch.flatten(resnet, start_dim=1)
        mlp = self.mlp(flat)
        if mode == 'train':
            proj_imgs = torch.ones_like(x)
            for i in range(mlp.shape[0]):
                mlp_ = quater2euler(mlp[i,:])
                proj_imgs[i] = torch_normalize(projectPDB_NP(pdb.to('cpu'), 128, 2.96, 1, mlp_[0].to('cpu'), mlp_[1].to('cpu'), mlp_[2].to('cpu'), 0, 0, 0))
            return mlp, proj_imgs.to('cuda:0')
        elif mode == 'inference':
            return mlp
    """
    def forward(self, x, mode = 'train'):
        resnet = self.resnet(x)
        flat = torch.flatten(resnet, start_dim=1)
        mlp = self.mlp(flat)
        return mlp
    
